# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
import json  # noqa: F401
import os
import random
import re
import time
import warnings
import weakref

from . import models, accounts, errors, utils
from .compat import six, Iterable, urlparse
from .config import options
from .errors import NoSuchObject, ODPSError
from .rest import RestClient
from .tempobj import clean_stored_objects


DEFAULT_ENDPOINT = 'http://service.odps.aliyun.com/api'
DEFAULT_REGION_NAME = 'cn'
LOGVIEW_HOST_DEFAULT = 'http://logview.aliyun.com'

_ALTER_TABLE_REGEX = re.compile(
    r'^\s*(drop|alter)\s+table\s*(|if\s+exists)\s+(?P<table_name>[^\s;]+)', re.I
)
_MERGE_SMALL_FILES_REGEX = re.compile(
    r'^alter\s+table\s+(?P<table>[^\s;]+)\s+'
    r'(|partition\s*\((?P<partition>[^)]+)\s*\))\s*'
    r'(merge\s+smallfiles|compact\s+(?P<compact_type>[^\s;]+))[\s;]*$',
    re.I,
)
_ENDPOINT_HOST_WITH_REGION_REGEX = re.compile(
    r'service\.([^\.]+)\.(odps|maxcompute)\.aliyun(|-inc)\.com', re.I
)

_logview_host_cache = dict()


@utils.attach_internal
class ODPS(object):
    """
    Main entrance to ODPS.

    Convenient operations on ODPS objects are provided.
    Please refer to `ODPS docs <https://help.aliyun.com/document_detail/27818.html>`_
    for more details.

    Generally, basic operations such as ``list``, ``get``, ``exist``, ``create``, ``delete``
    are provided for each ODPS object.
    Take the ``Table`` as an example.

    To create an ODPS instance, access_id and access_key is required, and should ensure correctness,
    or ``SignatureNotMatch`` error will throw. If `tunnel_endpoint` is not set, the tunnel API will
    route service URL automatically.

    :param access_id: Aliyun Access ID
    :param secret_access_key: Aliyun Access Key
    :param project: default project name
    :param endpoint: Rest service URL
    :param tunnel_endpoint: Tunnel service URL
    :param logview_host: Logview host URL
    :param app_account: Application account, instance of `odps.accounts.AppAccount` used for dual authentication

    :Example:

    >>> odps = ODPS('**your access id**', '**your access key**', 'default_project')
    >>>
    >>> for table in odps.list_tables():
    >>>    # handle each table
    >>>
    >>> table = odps.get_table('dual')
    >>>
    >>> odps.exist_table('dual') is True
    >>>
    >>> odps.create_table('test_table', schema)
    >>>
    >>> odps.delete_table('test_table')
    """
    def __init__(
        self, access_id=None, secret_access_key=None, project=None,
        endpoint=None, schema=None, app_account=None, logview_host=None,
        tunnel_endpoint=None, region_name=None, **kw
    ):
        # avoid polluted copy sources :(
        access_id = utils.strip_if_str(access_id)
        secret_access_key = utils.strip_if_str(secret_access_key)
        project = utils.strip_if_str(project)
        endpoint = utils.strip_if_str(endpoint)
        schema = utils.strip_if_str(schema)
        logview_host = utils.strip_if_str(logview_host)
        tunnel_endpoint = utils.strip_if_str(tunnel_endpoint)
        region_name = utils.strip_if_str(region_name)

        if isinstance(access_id, accounts.BaseAccount):
            assert secret_access_key is None, "Cannot supply secret_access_key with an account"
            kw["account"], access_id = access_id, None

        self._init(
            access_id=access_id, secret_access_key=secret_access_key,
            project=project, endpoint=endpoint, schema=schema,
            app_account=app_account, logview_host=logview_host,
            tunnel_endpoint=tunnel_endpoint, region_name=region_name, **kw
        )
        clean_stored_objects(self)

    def _init(
        self, access_id=None, secret_access_key=None, project=None,
        endpoint=None, schema=None, region_name=None, **kw
    ):
        self._property_update_callbacks = set()

        account = kw.pop('account', None)
        self.app_account = kw.pop('app_account', None)

        if account is None:
            if access_id is not None:
                self.account = self._build_account(access_id, secret_access_key)
            elif accounts.BearerTokenAccount.from_environments():
                self.account = accounts.BearerTokenAccount.from_environments()
            elif options.account is not None:
                self.account = options.account
            else:
                raise TypeError('`access_id` and `secret_access_key` should be provided.')
        else:
            self.account = account
        self.endpoint = endpoint or options.endpoint or DEFAULT_ENDPOINT
        self.project = project or options.default_project
        self.region_name = region_name or self._get_region_from_endpoint(self.endpoint)
        self._schema = schema
        self.rest = RestClient(
            self.account, self.endpoint, project, schema,
            app_account=self.app_account, proxy=options.api_proxy,
            region_name=self.region_name, tag="ODPS",
        )

        self._tunnel_endpoint = kw.pop('tunnel_endpoint', None) or options.tunnel.endpoint

        self._logview_host = (
            kw.pop("logview_host", None)
            or options.logview_host
            or self.get_logview_host()
        )

        self._default_tenant = models.Tenant(client=self.rest)

        self._projects = models.Projects(client=self.rest, _odps_ref=weakref.ref(self))
        if project:
            self._project = self.get_project()

        self._seahawks_url = kw.pop("seahawks_url", None)
        if self._seahawks_url:
            options.seahawks_url = self._seahawks_url

        self._default_session = None
        self._default_session_name = ""

        # Make instance to global
        overwrite_global = kw.pop("overwrite_global", True)
        if overwrite_global and options.is_global_account_overwritable:
            self.to_global(overwritable=True)

        if kw:
            raise TypeError(
                "Argument %s not acceptable, please check your spellings"
                % ", ".join(kw.keys()),
            )

    @staticmethod
    def _get_region_from_endpoint(endpoint):
        parsed = urlparse(endpoint)
        match = _ENDPOINT_HOST_WITH_REGION_REGEX.search(parsed.hostname or "")
        if match is None:
            return DEFAULT_REGION_NAME
        return match.group(1)

    def __getstate__(self):
        params = dict(
            project=self.project,
            endpoint=self.endpoint,
            tunnel_endpoint=self._tunnel_endpoint,
            logview_host=self._logview_host,
            schema=self.schema,
            seahawks_url=self._seahawks_url
        )
        if isinstance(self.account, accounts.AliyunAccount):
            params.update(dict(access_id=self.account.access_id,
                               secret_access_key=self.account.secret_access_key))
        return params

    def __setstate__(self, state):
        if 'secret_access_key' in state:
            if os.environ.get('ODPS_ENDPOINT', None) is not None:
                state['endpoint'] = os.environ['ODPS_ENDPOINT']
            self._init(**state)
            return

        bearer_token_account = accounts.BearerTokenAccount.from_environments()
        if bearer_token_account is not None:
            state['project'] = os.environ.get('ODPS_PROJECT_NAME')
            state['endpoint'] = os.environ.get('ODPS_RUNTIME_ENDPOINT') or os.environ['ODPS_ENDPOINT']
            state.pop('access_id', None)
            state.pop('secret_access_key', None)
            self._init(None, None, account=bearer_token_account, **state)
        else:
            self._init(**state)

    def as_account(self, access_id=None, secret_access_key=None, account=None, app_account=None):
        """
        Creates a new ODPS entry object with a new account information

        :param access_id: Aliyun Access ID of the new account
        :param secret_access_key: Aliyun Access Key of the new account
        :param account: new account object, if `access_id` and `secret_access_key` not supplied
        :param app_account: Application account, instance of `odps.accounts.AppAccount`
            used for dual authentication
        :return:
        """
        if access_id is not None and secret_access_key is not None:
            assert account is None
            account = accounts.AliyunAccount(access_id, secret_access_key)

        params = dict(
            project=self.project,
            endpoint=self.endpoint,
            tunnel_endpoint=self._tunnel_endpoint,
            logview_host=self._logview_host,
            schema=self.schema,
            seahawks_url=self._seahawks_url,
            account=account or self.account,
            app_account=app_account or self.app_account,
            overwrite_global=False,
        )
        return ODPS(**params)

    def __mars_tokenize__(self):
        return self.__getstate__()

    @classmethod
    def _from_account(cls, account, project, endpoint=DEFAULT_ENDPOINT,
                      tunnel_endpoint=None, **kwargs):
        return cls(None, None, project, endpoint=endpoint,
                   tunnel_endpoint=tunnel_endpoint, account=account, **kwargs)

    @property
    def default_tenant(self):
        return self._default_tenant

    def is_schema_namespace_enabled(self, settings=None):
        settings = settings or {}
        setting = str(
            settings.get("odps.namespace.schema")
            or (options.sql.settings or {}).get("odps.namespace.schema")
            or ("true" if options.always_enable_schema else None)
            or self.default_tenant.get_parameter("odps.namespace.schema")
            or "false"
        )
        return setting.lower() == "true"

    @property
    def projects(self):
        return self._projects

    @property
    def schema(self):
        """
        Get or set default schema name of the ODPS object
        """
        default_schema = "default" if self.is_schema_namespace_enabled() else None
        return self._schema or default_schema

    @schema.setter
    def schema(self, value):
        self._schema = value
        for cb in self._property_update_callbacks:
            cb(self)

    @property
    def tunnel_endpoint(self):
        """
        Get or set tunnel endpoint of the ODPS object
        """
        return self._tunnel_endpoint

    @tunnel_endpoint.setter
    def tunnel_endpoint(self, value):
        self._tunnel_endpoint = value
        for cb in self._property_update_callbacks:
            cb(self)

    def list_projects(
        self, owner=None, user=None, group=None, prefix=None, max_items=None,
        region_id=None, tenant_id=None
    ):
        """
        List projects.

        :param owner: Aliyun account, the owner which listed projects belong to
        :param user: name of the user who has access to listed projects
        :param group: name of the group listed projects belong to
        :param prefix: prefix of names of listed projects
        :param max_items: the maximal size of result set
        :return: projects in this endpoint.
        :rtype: generator
        """
        return self.projects.iterate(
            owner=owner, user=user, group=group, max_items=max_items, name=prefix,
            region_id=region_id, tenant_id=tenant_id
        )

    @property
    def logview_host(self):
        return self._logview_host

    def get_project(self, name=None, default_schema=None):
        """
        Get project by given name.

        :param name: project name, if not provided, will be the default project
        :param default_schema: default schema name, if not provided, will be
            the schema specified in ODPS object
        :return: the right project
        :rtype: :class:`odps.models.Project`
        :raise: :class:`odps.errors.NoSuchObject` if not exists

        .. seealso:: :class:`odps.models.Project`
        """

        if name is None:
            name = self.project
        elif isinstance(name, models.Project):
            return name
        proj = self._projects[name]
        proj._tunnel_endpoint = self._tunnel_endpoint
        proj._logview_host = self._logview_host
        # use _schema to avoid requesting for tenant options
        proj._default_schema = default_schema or self._schema

        proj_ref = weakref.ref(proj)

        def project_update_callback(odps, update_schema=True):
            proj_obj = proj_ref()
            if proj_obj:
                if update_schema:
                    proj_obj._default_schema = odps.schema
                proj_obj._tunnel_endpoint = odps.tunnel_endpoint
            else:
                self._property_update_callbacks.difference_update([project_update_callback])

        # we need to update default schema value on the project
        self._property_update_callbacks.add(
            functools.partial(project_update_callback, update_schema=default_schema is None)
        )
        return proj

    def exist_project(self, name):
        """
        If project name which provided exists or not.

        :param name: project name
        :return: True if exists or False
        :rtype: bool
        """

        return name in self._projects

    def list_schemas(self, project=None, prefix=None, owner=None):
        """
        List all schemas of a project.

        :param project: project name, if not provided, will be the default project
        :param str prefix: the listed schemas start with this **prefix**
        :param str owner: Aliyun account, the owner which listed tables belong to
        :return: schemas
        """
        project = self.get_project(name=project)
        return project.schemas.iterate(name=prefix, owner=owner)

    def get_schema(self, name=None, project=None):
        """
        Get the schema by given name.

        :param name: schema name, if not provided, will be the default schema
        :param project: project name, if not provided, will be the default project
        :return: the Schema object
        """
        project = self.get_project(name=project)
        return project.schemas[name or self.schema]

    def exist_schema(self, name, project=None):
        """
        If schema name which provided exists or not.

        :param name: schema name
        :param project: project name, if not provided, will be the default project
        :return: True if exists or False
        :rtype: bool
        """
        project = self.get_project(name=project)
        return name in project.schemas

    @utils.with_wait_argument
    def create_schema(self, name, project=None, async_=False):
        """
        Create a schema with given name

        :param name: schema name
        :param project: project name, if not provided, will be the default project
        :param async_: if True, will run asynchronously
        :return: if async_ is True, return instance, otherwise return Schema object.
        """
        project = self.get_project(name=project)
        return project.schemas.create(name, async_=async_)

    @utils.with_wait_argument
    def delete_schema(self, name, project=None, async_=False):
        """
        Delete the schema with given name

        :param name: schema name
        :param project: project name, if not provided, will be the default project
        :param async_: if True, will run asynchronously
        :type async_: bool
        """
        project = self.get_project(name=project)
        return project.schemas.delete(name, async_=async_)

    def _get_project_or_schema(self, project=None, schema=None):
        if self.is_schema_namespace_enabled():
            schema = schema or "default"
        if schema is not None:
            return self.get_schema(schema, project=project)
        else:
            return self.get_project(project)

    def _split_object_dots(self, name):
        parts = [x.strip() for x in name.split('.')]
        if len(parts) == 1:
            project, schema, name = None, None, parts[0]
        elif len(parts) == 2:
            if self.is_schema_namespace_enabled():
                schema, name = parts
                project = None
            else:
                project, name = parts
                schema = None
        else:
            project, schema, name = parts
        if name.startswith("`") and name.endswith("`"):
            name = name.strip("`")
        return project, schema, name

    def list_tables(
        self, project=None, prefix=None, owner=None, schema=None, type=None, extended=False
    ):
        """
        List all tables of a project.
        If prefix is provided, the listed tables will all start with this prefix.
        If owner is provided, the listed tables will belong to such owner.

        :param str project: project name, if not provided, will be the default project
        :param str prefix: the listed tables start with this **prefix**
        :param str owner: Aliyun account, the owner which listed tables belong to
        :param str schema: schema name, if not provided, will be the default schema
        :param str type: type of the table
        :param bool extended: if True, load extended information for table
        :return: tables in this project, filtered by the optional prefix and owner.
        :rtype: generator
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.tables.iterate(name=prefix, owner=owner, type=type, extended=extended)

    def get_table(self, name, project=None, schema=None):
        """
        Get table by given name.

        :param name: table name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: the right table
        :rtype: :class:`odps.models.Table`
        :raise: :class:`odps.errors.NoSuchObject` if not exists

        .. seealso:: :class:`odps.models.Table`
        """

        if isinstance(name, six.string_types) and '.' in name:
            project, schema, name = self._split_object_dots(name)

        parent = self._get_project_or_schema(project, schema)
        return parent.tables[name]

    def exist_table(self, name, project=None, schema=None):
        """
        If the table with given name exists or not.

        :param name: table name
        :param project: project name, if not provided, will be the default project
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :return: True if table exists or False
        :rtype: bool
        """

        if isinstance(name, six.string_types) and '.' in name:
            project, schema, name = self._split_object_dots(name)

        parent = self._get_project_or_schema(project, schema)
        return name in parent.tables

    @utils.with_wait_argument
    def create_table(
        self, name, table_schema=None, project=None, schema=None, comment=None,
        if_not_exists=False, lifecycle=None, shard_num=None, hub_lifecycle=None,
        hints=None, transactional=False, primary_key=None, storage_tier=None,
        async_=False, **kw
    ):
        """
        Create a table by given schema and other optional parameters.

        :param name: table name
        :param table_schema: table schema. Can be an instance
            of :class:`odps.models.TableSchema` or a string like 'col1 string, col2 bigint'
        :param project: project name, if not provided, will be the default project
        :param comment: table comment
        :param str schema: schema name, if not provided, will be the default schema
        :param bool if_not_exists: will not create if this table already exists, default False
        :param int lifecycle: table's lifecycle. If absent, `options.lifecycle` will be used.
        :param int shard_num: table's shard num
        :param int hub_lifecycle: hub lifecycle
        :param dict hints: hints for the task
        :param bool transactional: make table transactional
        :param list primary_key: primary key of the table, only for transactional tables
        :param str storage_tier: storage tier of the table
        :param bool async_: if True, will run asynchronously
        :return: the created Table if not async else odps instance
        :rtype: :class:`odps.models.Table` or :class:`odps.models.Instance`

        .. seealso:: :class:`odps.models.Table`, :class:`odps.models.TableSchema`
        """
        from .types import OdpsSchema

        if table_schema is None and schema:
            if (
                isinstance(schema, OdpsSchema)
                or isinstance(schema, tuple)
                or (isinstance(schema, six.string_types) and ' ' in schema)
            ):
                table_schema, schema = schema, None
                warnings.warn(
                    "`schema` is renamed as `table_schema` in `create_table`, "
                    "the original parameter now represents schema name. Please "
                    "change your code.",
                    DeprecationWarning,
                    stacklevel=2
                )
                utils.add_survey_call("ODPS.create_table(schema='schema_name')")

        if table_schema is None:
            raise TypeError("`table_schema` argument not filled")

        if isinstance(name, six.string_types) and '.' in name:
            project, schema, name = self._split_object_dots(name)

        if lifecycle is None and options.lifecycle is not None:
            lifecycle = options.lifecycle

        parent = self._get_project_or_schema(project, schema)
        return parent.tables.create(
            name, table_schema, comment=comment, if_not_exists=if_not_exists,
            lifecycle=lifecycle, shard_num=shard_num, hub_lifecycle=hub_lifecycle,
            hints=hints, transactional=transactional, primary_key=primary_key,
            storage_tier=storage_tier, async_=async_, **kw
        )

    @utils.with_wait_argument
    def delete_table(self, name, project=None, if_exists=False, schema=None, hints=None, async_=False):
        """
        Delete the table with given name

        :param name: table name
        :param project: project name, if not provided, will be the default project
        :param bool if_exists:  will not raise errors when the table does not exist, default False
        :param str schema: schema name, if not provided, will be the default schema
        :param dict hints: hints for the task
        :param bool async_: if True, will run asynchronously
        :return: None if not async else odps instance
        """

        if isinstance(name, six.string_types) and '.' in name:
            project, schema, name = self._split_object_dots(name)

        parent = self._get_project_or_schema(project, schema)
        return parent.tables.delete(name, if_exists=if_exists, hints=hints, async_=async_)

    def read_table(self, name, limit=None, start=0, step=None,
                   project=None, schema=None, partition=None, **kw):
        """
        Read table's records.

        :param name: table or table name
        :type name: :class:`odps.models.table.Table` or str
        :param limit:  the records' size, if None will read all records from the table
        :param start:  the record where read starts with
        :param step:  default as 1
        :param project: project name, if not provided, will be the default project
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :param partition: the partition of this table to read
        :param columns: the columns' names which are the parts of table's columns
        :type columns: list
        :param compress: if True, the data will be compressed during downloading
        :type compress: bool
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :param endpoint: tunnel service URL
        :param reopen: reading the table will reuse the session which opened last time,
                       if set to True will open a new download session, default as False
        :return: records
        :rtype: generator

        :Example:

        >>> for record in odps.read_table('test_table', 100):
        >>>     # deal with such 100 records
        >>> for record in odps.read_table('test_table', partition='pt=test', start=100, limit=100):
        >>>     # read the `pt=test` partition, skip 100 records and read 100 records

        .. seealso:: :class:`odps.models.Record`
        """

        if isinstance(name, six.string_types) and '.' in name:
            project, schema, name = self._split_object_dots(name)

        if not isinstance(name, six.string_types):
            if name.get_schema():
                schema = name.get_schema().name
            project, name = name.project.name, name.name

        parent = self._get_project_or_schema(project, schema)
        table = parent.tables[name]

        compress = kw.pop('compress', False)
        columns = kw.pop('columns', None)

        with table.open_reader(partition=partition, **kw) as reader:
            for record in reader.read(start, limit, step=step,
                                      compress=compress, columns=columns):
                yield record

    def write_table(self, name, *block_records, **kw):
        """
        Write records into given table.

        :param name: table or table name
        :type name: :class:`.models.table.Table` or str
        :param block_records: if given records only, the block id will be 0 as default.
        :param project: project name, if not provided, will be the default project
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :param partition: the partition of this table to write
        :param bool overwrite: if True, will overwrite existing data
        :param compress: if True, the data will be compressed during uploading
        :type compress: bool
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :param endpoint:  tunnel service URL
        :param reopen: writing the table will reuse the session which opened last time,
                       if set to True will open a new upload session, default as False
        :return: None

        :Example:

        >>> odps.write_table('test_table', records)  # write to block 0 as default
        >>>
        >>> odps.write_table('test_table', 0, records)  # write to block 0 explicitly
        >>>
        >>> odps.write_table('test_table', 0, records1, 1, records2)  # write to multi-blocks
        >>>
        >>> odps.write_table('test_table', records, partition='pt=test') # write to certain partition

        .. seealso:: :class:`odps.models.Record`
        """

        project = kw.pop('project', None)
        schema = kw.pop('schema', None)

        if isinstance(name, six.string_types) and '.' in name:
            project, schema, name = self._split_object_dots(name)

        if not isinstance(name, six.string_types):
            if name.get_schema():
                schema = name.get_schema().name
            project, name = name.project.name, name.name

        parent = self._get_project_or_schema(project, schema)
        table = parent.tables[name]
        partition = kw.pop('partition', None)

        if len(block_records) == 1 and isinstance(block_records[0], Iterable):
            blocks = [0, ]
            records_iterators = block_records
        else:
            blocks = block_records[::2]
            records_iterators = block_records[1::2]

            if len(blocks) != len(records_iterators):
                raise ValueError('Should invoke like '
                                 'odps.write_table(block_id, records, block_id2, records2, ..., **kw)')

        with table.open_writer(partition=partition, blocks=blocks, **kw) as writer:
            for block, records in zip(blocks, records_iterators):
                writer.write(block, records)

    def list_resources(self, project=None, prefix=None, owner=None, schema=None):
        """
        List all resources of a project.

        :param project: project name, if not provided, will be the default project
        :param str prefix: the listed resources start with this **prefix**
        :param str owner: Aliyun account, the owner which listed tables belong to
        :param str schema: schema name, if not provided, will be the default schema
        :return: resources
        :rtype: generator
        """

        parent = self._get_project_or_schema(project, schema)
        for resource in parent.resources.iterate(name=prefix, owner=owner):
            yield resource

    def get_resource(self, name, project=None, schema=None):
        """
        Get a resource by given name

        :param name: resource name
        :param project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: the right resource
        :rtype: :class:`odps.models.Resource`
        :raise: :class:`odps.errors.NoSuchObject` if not exists

        .. seealso:: :class:`odps.models.Resource`
        """

        parent = self._get_project_or_schema(project, schema)
        return parent.resources[name]

    def exist_resource(self, name, project=None, schema=None):
        """
        If the resource with given name exists or not.

        :param name: resource name
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :param project: project name, if not provided, will be the default project
        :return: True if exists or False
        :rtype: bool
        """

        parent = self._get_project_or_schema(project, schema)
        return name in parent.resources

    def open_resource(
        self, name, project=None, mode='r+', encoding='utf-8', schema=None,
        type="file", stream=False, comment=None, temp=False
    ):
        """
        Open a file resource as file-like object.
        This is an elegant and pythonic way to handle file resource.

        The argument ``mode`` stands for the open mode for this file resource.
        It can be binary mode if the 'b' is inside. For instance,
        'rb' means opening the resource as read binary mode
        while 'r+b' means opening the resource as read+write binary mode.
        This is most import when the file is actually binary such as tar or jpeg file,
        so be aware of opening this file as a correct mode.

        Basically, the text mode can be 'r', 'w', 'a', 'r+', 'w+', 'a+'
        just like the builtin python ``open`` method.

        * ``r`` means read only
        * ``w`` means write only, the file will be truncated when opening
        * ``a`` means append only
        * ``r+`` means read+write without constraint
        * ``w+`` will truncate first then opening into read+write
        * ``a+`` can read+write, however the written content can only be appended to the end

        :param name: file resource or file resource name
        :type name: :class:`odps.models.FileResource` or str
        :param project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param str mode: the mode of opening file, described as above
        :param str encoding: utf-8 as default
        :param str type: resource type, can be "file", "archive", "jar" or "py"
        :param bool stream: if True, use stream to upload, False by default
        :param str comment: comment of the resource
        :return: file-like object

        :Example:

        >>> with odps.open_resource('test_resource', mode='r') as fp:
        >>>     fp.read(1)  # read one unicode character
        >>>     fp.write('test')  # wrong, cannot write under read mode
        >>>
        >>> with odps.open_resource('test_resource', mode='wb') as fp:
        >>>     fp.readlines() # wrong, cannot read under write mode
        >>>     fp.write('hello world') # write bytes
        >>>
        >>> with odps.open_resource('test_resource') as fp: # default as read-write mode
        >>>     fp.seek(5)
        >>>     fp.truncate()
        >>>     fp.flush()
        """

        from .models import FileResource

        if isinstance(name, FileResource):
            return name.open(mode=mode)

        parent = self._get_project_or_schema(project, schema)
        return parent.resources.get_typed(name, type=type, comment=comment, temp=temp).open(
            mode=mode, encoding=encoding, stream=stream
        )

    def create_resource(self, name, type=None, project=None, schema=None, **kwargs):
        """
        Create a resource by given name and given type.

        Currently, the resource type can be ``file``, ``jar``, ``py``, ``archive``, ``table``.

        The ``file``, ``jar``, ``py``, ``archive`` can be classified into file resource.
        To init the file resource, you have to provide another parameter which is a file-like object.

        For the table resource, the table name, project name, and partition should be provided
        which the partition is optional.

        :param name: resource name
        :param type: resource type, now support ``file``, ``jar``, ``py``, ``archive``, ``table``
        :param project: project name, if not provided, will be the default project
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :param kwargs: optional arguments, I will illustrate this in the example below.
        :return: resource depends on the type, if ``file`` will be :class:`odps.models.FileResource` and so on
        :rtype: :class:`odps.models.Resource`'s subclasses

        :Example:

        >>> from odps.models.resource import *
        >>>
        >>> res = odps.create_resource('test_file_resource', 'file', fileobj=open('/to/path/file'))
        >>> assert isinstance(res, FileResource)
        >>> True
        >>>
        >>> res = odps.create_resource('test_py_resource.py', 'py', fileobj=StringIO('import this'))
        >>> assert isinstance(res, PyResource)
        >>> True
        >>>
        >>> res = odps.create_resource('test_table_resource', 'table', table_name='test_table', partition='pt=test')
        >>> assert isinstance(res, TableResource)
        >>> True
        >>>

        .. seealso:: :class:`odps.models.FileResource`, :class:`odps.models.PyResource`,
                     :class:`odps.models.JarResource`, :class:`odps.models.ArchiveResource`,
                     :class:`odps.models.TableResource`
        """

        type_ = kwargs.get('typo') or type
        parent = self._get_project_or_schema(project, schema)
        return parent.resources.create(name=name, type=type_, **kwargs)

    def delete_resource(self, name, project=None, schema=None):
        """
        Delete resource by given name.

        :param name: resource name
        :param project: project name, if not provided, will be the default project
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :return: None
        """

        parent = self._get_project_or_schema(project, schema)
        return parent.resources.delete(name)

    def list_functions(self, project=None, prefix=None, owner=None, schema=None):
        """
        List all functions of a project.

        :param str project: project name, if not provided, will be the default project
        :param str prefix: the listed functions start with this **prefix**
        :param str owner: Aliyun account, the owner which listed tables belong to
        :param str schema: schema name, if not provided, will be the default schema
        :return: functions
        :rtype: generator
        """

        parent = self._get_project_or_schema(project, schema)
        for function in parent.functions.iterate(name=prefix, owner=owner):
            yield function

    def get_function(self, name, project=None, schema=None):
        """
        Get the function by given name

        :param name: function name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: the right function
        :raise: :class:`odps.errors.NoSuchObject` if not exists

        .. seealso:: :class:`odps.models.Function`
        """

        parent = self._get_project_or_schema(project, schema)
        return parent.functions[name]

    def exist_function(self, name, project=None, schema=None):
        """
        If the function with given name exists or not.

        :param str name: function name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: True if the function exists or False
        :rtype: bool
        """

        parent = self._get_project_or_schema(project, schema)
        return name in parent.functions

    def create_function(self, name, project=None, schema=None, **kwargs):
        """
        Create a function by given name.

        :param name: function name
        :param project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param str class_type: main class
        :param list resources: the resources that function needs to use
        :return: the created function
        :rtype: :class:`odps.models.Function`

        :Example:

        >>> res = odps.get_resource('test_func.py')
        >>> func = odps.create_function('test_func', class_type='test_func.Test', resources=[res, ])

        .. seealso:: :class:`odps.models.Function`
        """

        parent = self._get_project_or_schema(project, schema)
        return parent.functions.create(name=name, **kwargs)

    def delete_function(self, name, project=None, schema=None):
        """
        Delete a function by given name.

        :param name: function name
        :param project: project name, if not provided, will be the default project
        :param schema: schema name, if not provided, will be the default schema
        :type schema: str
        :return: None
        """

        parent = self._get_project_or_schema(project, schema)
        return parent.functions.delete(name)

    def list_instances(self, project=None, start_time=None, end_time=None,
                       status=None, only_owner=None, quota_index=None, **kw):
        """
        List instances of a project by given optional conditions
        including start time, end time, status and if only the owner.

        :param project: project name, if not provided, will be the default project
        :param start_time: the start time of filtered instances
        :type start_time: datetime, int or float
        :param end_time: the end time of filtered instances
        :type end_time: datetime, int or float
        :param status: including 'Running', 'Suspended', 'Terminated'
        :param only_owner: True will filter the instances created by current user
        :type only_owner: bool
        :param quota_index:
        :type quota_index: str
        :return: instances
        :rtype: list
        """
        if 'from_time' in kw:
            start_time = kw['from_time']
            warnings.warn(
                'The keyword argument `from_time` has been replaced by `start_time`.',
                DeprecationWarning,
            )

        project = self.get_project(name=project)
        return project.instances.iterate(
            start_time=start_time, end_time=end_time,
            status=status, only_owner=only_owner, quota_index=quota_index)

    def list_instance_queueing_infos(self, project=None, status=None, only_owner=None,
                                     quota_index=None):
        """
        List instance queueing information.

        :param project: project name, if not provided, will be the default project
        :param status: including 'Running', 'Suspended', 'Terminated'
        :param only_owner: True will filter the instances created by current user
        :type only_owner: bool
        :param quota_index:
        :type quota_index: str
        :return: instance queueing infos
        :rtype: list
        """

        project = self.get_project(name=project)
        return project.instance_queueing_infos.iterate(
            status=status, only_owner=only_owner, quota_index=quota_index)

    def get_instance(self, id_, project=None):
        """
        Get instance by given instance id.

        :param id_: instance id
        :param project: project name, if not provided, will be the default project
        :return: the right instance
        :rtype: :class:`odps.models.Instance`
        :raise: :class:`odps.errors.NoSuchObject` if not exists

        .. seealso:: :class:`odps.models.Instance`
        """

        project = self.get_project(name=project)
        return project.instances[id_]

    def exist_instance(self, id_, project=None):
        """
        If the instance with given id exists or not.

        :param id_: instance id
        :param project: project name, if not provided, will be the default project
        :return: True if exists or False
        :rtype: bool
        """

        project = self.get_project(name=project)
        return id_ in project.instances

    def stop_instance(self, id_, project=None):
        """
        Stop the running instance by given instance id.

        :param id_: instance id
        :param project: project name, if not provided, will be the default project
        :return: None
        """

        project = self.get_project(name=project)
        project.instances[id_].stop()

    stop_job = stop_instance  # to keep compatible

    def execute_sql(self, sql, project=None, priority=None, running_cluster=None,
                    hints=None, **kwargs):
        """
        Run a given SQL statement and block until the SQL executed successfully.

        :param sql: SQL statement
        :type sql: str
        :param project: project name, if not provided, will be the default project
        :param priority: instance priority, 9 as default
        :type priority: int
        :param running_cluster: cluster to run this instance
        :param hints: settings for SQL, e.g. `odps.mapred.map.split.size`
        :type hints: dict
        :return: instance
        :rtype: :class:`odps.models.Instance`

        :Example:

        >>> instance = odps.execute_sql('select * from dual')
        >>> with instance.open_reader() as reader:
        >>>     for record in reader:  # iterate to handle result with schema
        >>>         # handle each record
        >>>
        >>> instance = odps.execute_sql('desc dual')
        >>> with instance.open_reader() as reader:
        >>>     print(reader.raw)  # without schema, just get the raw result

        .. seealso:: :class:`odps.models.Instance`
        """

        async_ = kwargs.pop('async_', kwargs.pop('async', False))

        inst = self.run_sql(
            sql, project=project, priority=priority, running_cluster=running_cluster,
            hints=hints, **kwargs)
        if not async_:
            inst.wait_for_success()
        return inst

    def run_sql(self, sql, project=None, priority=None, running_cluster=None,
                hints=None, aliases=None, default_schema=None, **kwargs):
        """
        Run a given SQL statement asynchronously

        :param sql: SQL statement
        :type sql: str
        :param project: project name, if not provided, will be the default project
        :param priority: instance priority, 9 as default
        :type priority: int
        :param running_cluster: cluster to run this instance
        :param hints: settings for SQL, e.g. `odps.mapred.map.split.size`
        :type hints: dict
        :param aliases:
        :type aliases: dict
        :return: instance
        :rtype: :class:`odps.models.Instance`

        .. seealso:: :class:`odps.models.Instance`
        """
        sql = utils.to_text(sql)

        merge_small_files_match = _MERGE_SMALL_FILES_REGEX.match(sql)
        if merge_small_files_match:
            kwargs = merge_small_files_match.groupdict().copy()
            kwargs.update({
                "project": project,
                "schema": default_schema,
                "hints": hints,
                "running_cluster": running_cluster,
                "priority": priority,
            })
            return self.run_merge_files(**kwargs)

        priority = priority if priority is not None else options.priority
        if priority is None and options.get_priority is not None:
            priority = options.get_priority(self)
        on_instance_create = kwargs.pop('on_instance_create', None)

        alter_table_match = _ALTER_TABLE_REGEX.match(sql)
        if alter_table_match:
            drop_table_name = alter_table_match.group('table_name')
            sql_project, sql_schema, sql_name = self._split_object_dots(drop_table_name)
            sql_project = sql_project or project
            sql_schema = sql_schema or default_schema
            del self._get_project_or_schema(sql_project, sql_schema).tables[sql_name]

        task = models.SQLTask(query=sql, **kwargs)
        task.update_sql_settings(hints)

        default_schema = default_schema or self.schema
        if default_schema is not None:
            task.update_sql_settings({
                "odps.sql.allow.namespace.schema": "true",
                "odps.namespace.schema": "true",
                "odps.default.schema": default_schema,
            })

        if aliases:
            task.update_aliases(aliases)

        project = self.get_project(name=project)
        try:
            return project.instances.create(
                task=task, priority=priority, running_cluster=running_cluster,
                create_callback=on_instance_create
            )
        except errors.ParseError as ex:
            ex.statement = sql
            raise

    def execute_sql_cost(self, sql, project=None, hints=None, **kwargs):
        """

        :param sql: SQL statement
        :type sql: str
        :param project: project name, if not provided, will be the default project
        :param hints: settings for SQL, e.g. `odps.mapred.map.split.size`
        :type hints: dict
        :return: cost info in dict format
        :rtype: cost: dict

        :Example:

        >>> sql_cost = odps.execute_sql_cost('select * from dual')
        >>> sql_cost.udf_num
        0
        >>> sql_cost.complexity
        1.0
        >>> sql_cost.input_size
        100

        """
        task = models.SQLCostTask(query=utils.to_text(sql), **kwargs)
        task.update_sql_cost_settings(hints)
        project = self.get_project(name=project)
        inst = project.instances.create(task=task)
        inst.wait_for_success()
        return inst.get_sql_task_cost()

    def _parse_partition_string(self, partition):
        parts = []
        for p in utils.split_quoted(partition, ','):
            kv = [pp.strip() for pp in utils.split_quoted(p, '=')]
            if len(kv) != 2:
                raise ValueError('Partition representation malformed.')
            if not kv[1].startswith('"') and not kv[1].startswith("'"):
                kv[1] = repr(kv[1])
            parts.append('%s=%s' % tuple(kv))
        return ','.join(parts)

    def execute_merge_files(self, table, partition=None, project=None, schema=None,
                            hints=None, priority=None, running_cluster=None, compact_type=None):
        """
        Execute a task to merge multiple files in tables and wait for termination.

        :param table: name of the table to optimize
        :param partition: partition to optimize
        :param project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param hints: settings for merge task.
        :param priority: instance priority, 9 as default
        :param running_cluster: cluster to run this instance
        :param compact_type: compact option for transactional table, can be major or minor.
        :return: instance
        :rtype: :class:`odps.models.Instance`
        """
        inst = self.run_merge_files(
            table, partition=partition, project=project, hints=hints,
            schema=schema, priority=priority, running_cluster=running_cluster,
            compact_type=compact_type,
        )
        inst.wait_for_success()
        return inst

    def run_merge_files(self, table, partition=None, project=None, schema=None, hints=None,
                        priority=None, running_cluster=None, compact_type=None):
        """
        Start running a task to merge multiple files in tables.

        :param table: name of the table to optimize
        :param partition: partition to optimize
        :param project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param hints: settings for merge task.
        :param priority: instance priority, 9 as default
        :param running_cluster: cluster to run this instance
        :param compact_type: compact option for transactional table, can be major or minor.
        :return: instance
        :rtype: :class:`odps.models.Instance`
        """
        from .models.tasks import MergeTask

        schema = schema or self.schema
        if not isinstance(table, six.string_types):
            if table.get_schema():
                schema = table.get_schema().name
            table_name = table.full_table_name
        else:
            table_name = table
            _, schema, _ = self._split_object_dots(table)

        if partition:
            table_name += " partition(%s)" % (self._parse_partition_string(partition))

        priority = priority if priority is not None else options.priority
        if priority is None and options.get_priority is not None:
            priority = options.get_priority(self)

        hints = hints or dict()
        if options.default_task_settings:
            hints.update(options.default_task_settings)

        if compact_type:
            hints.update({
                "odps.merge.txn.table.compact": compact_type,
                "odps.merge.restructure.action": "hardlink",
            })
        if schema is not None:
            hints.update({
                "odps.sql.allow.namespace.schema": "true",
                "odps.namespace.schema": "true",
                "odps.default.schema": schema,
            })

        task = MergeTask(table=table_name.replace("`", ""))
        task.update_settings(hints)

        project = self.get_project(name=project)
        return project.instances.create(
            task=task, running_cluster=running_cluster, priority=priority
        )

    def execute_archive_table(self, table, partition=None, project=None, schema=None,
                              hints=None, priority=None):
        """
        Execute a task to archive tables and wait for termination.

        :param table: name of the table to archive
        :param partition: partition to archive
        :param project: project name, if not provided, will be the default project
        :param hints: settings for table archive task.
        :param priority: instance priority, 9 as default
        :return: instance
        :rtype: :class:`odps.models.Instance`
        """
        inst = self.run_archive_table(table, partition=partition, project=project, hints=hints,
                                      priority=priority)
        inst.wait_for_success()
        return inst

    def run_archive_table(self, table, partition=None, project=None, schema=None,
                          hints=None, priority=None):
        """
        Start running a task to archive tables.

        :param table: name of the table to archive
        :param partition: partition to archive
        :param project: project name, if not provided, will be the default project
        :param hints: settings for table archive task.
        :param priority: instance priority, 9 as default
        :return: instance
        :rtype: :class:`odps.models.Instance`
        """
        from .models.tasks import MergeTask

        if partition:
            table += " partition(%s)" % (self._parse_partition_string(partition))

        priority = priority if priority is not None else options.priority
        if priority is None and options.get_priority is not None:
            priority = options.get_priority(self)

        name = 'archive_task_{0}_{1}'.format(int(time.time()), random.randint(100000, 999999))
        task = MergeTask(name=name, table=table)
        task.update_settings(hints)
        task._update_property_json('archiveSettings', {'odps.merge.archive.flag': True})

        project = self.get_project(name=project)
        return project.instances.create(task=task, priority=priority)

    def list_volumes(self, project=None, schema=None, owner=None):
        """
        List volumes of a project.

        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param str owner: Aliyun account
        :return: volumes
        :rtype: list
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.volumes.iterate(owner=owner)

    @utils.deprecated('`create_volume` is deprecated. Use `created_parted_volume` instead.')
    def create_volume(self, name, project=None, **kwargs):
        self.create_parted_volume(name, project=project, **kwargs)

    def create_parted_volume(self, name, project=None, schema=None, **kwargs):
        """
        Create an old-fashioned partitioned volume in a project.

        :param str name: volume name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: volume
        :rtype: :class:`odps.models.PartedVolume`

        .. seealso:: :class:`odps.models.PartedVolume`
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.volumes.create_parted(name=name, **kwargs)

    def create_fs_volume(self, name, project=None, schema=None, **kwargs):
        """
        Create a new-fashioned file system volume in a project.

        :param str name: volume name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: volume
        :rtype: :class:`odps.models.FSVolume`

        .. seealso:: :class:`odps.models.FSVolume`
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.volumes.create_fs(name=name, **kwargs)

    def create_external_volume(
        self, name, project=None, schema=None, location=None, rolearn=None, **kwargs
    ):
        """
        Create a file system volume based on external storage (for instance, OSS) in a project.

        :param str name: volume name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: volume
        :rtype: :class:`odps.models.FSVolume`

        .. seealso:: :class:`odps.models.FSVolume`
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.volumes.create_external(
            name=name, location=location, rolearn=rolearn, **kwargs
        )

    def exist_volume(self, name, schema=None, project=None):
        """
        If the volume with given name exists or not.

        :param str name: volume name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: True if exists or False
        :rtype: bool
        """
        parent = self._get_project_or_schema(project, schema)
        return name in parent.volumes

    def get_volume(self, name, project=None, schema=None):
        """
        Get volume by given name.

        :param str name: volume name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: volume object. Return type depends on the type of the volume.
        :rtype: :class:`odps.models.Volume`
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.volumes[name]

    def delete_volume(self, name, project=None, schema=None):
        """
        Delete volume by given name.

        :param name: volume name
        :param project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: None
        """
        parent = self._get_project_or_schema(project, schema)
        return parent.volumes.delete(name)

    def list_volume_partitions(self, volume, project=None, schema=None):
        """
        List partitions of a volume.

        :param str volume: volume name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: partitions
        :rtype: list
        """
        volume = self.get_volume(volume, project, schema=schema)
        return volume.partitions.iterate()

    def get_volume_partition(self, volume, partition=None, project=None, schema=None):
        """
        Get partition in a parted volume by given name.

        :param str volume: volume name
        :param str partition: partition name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: partitions
        :rtype: :class:`odps.models.VolumePartition`
        """
        if partition is None:
            if not volume.startswith('/') or '/' not in volume.lstrip('/'):
                raise ValueError('You should provide a partition name or use partition path instead.')
            volume, partition = volume.lstrip('/').split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        return volume.partitions[partition]

    def exist_volume_partition(self, volume, partition=None, project=None, schema=None):
        """
        If the volume with given name exists in a partition or not.

        :param str volume: volume name
        :param str partition: partition name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        """
        if partition is None:
            if not volume.startswith('/') or '/' not in volume.lstrip('/'):
                raise ValueError('You should provide a partition name or use partition path instead.')
            volume, partition = volume.lstrip('/').split('/', 1)
        try:
            volume = self.get_volume(volume, project, schema=schema)
        except errors.NoSuchObject:
            return False
        return partition in volume.partitions

    def delete_volume_partition(self, volume, partition=None, project=None, schema=None):
        """
        Delete partition in a volume by given name

        :param str volume: volume name
        :param str partition: partition name
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        """
        if partition is None:
            if not volume.startswith('/') or '/' not in volume.lstrip('/'):
                raise ValueError('You should provide a partition name or use partition path instead.')
            volume, partition = volume.lstrip('/').split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        return volume.delete_partition(partition)

    def list_volume_files(self, volume, partition=None, project=None, schema=None):
        """
        List files in a volume. In partitioned volumes, the function returns files under specified partition.
        In file system volumes, the function returns files under specified path.

        :param str volume: volume name
        :param str partition: partition name for partitioned volumes, and path for file system volumes.
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :return: files
        :rtype: list

        :Example:
        >>> # List files under a partition in a partitioned volume. Two calls are equivalent.
        >>> odps.list_volume_files('parted_volume', 'partition_name')
        >>> odps.list_volume_files('/parted_volume/partition_name')
        >>> # List files under a path in a file system volume. Two calls are equivalent.
        >>> odps.list_volume_files('fs_volume', 'dir1/dir2')
        >>> odps.list_volume_files('/fs_volume/dir1/dir2')
        """
        from .models import PartedVolume
        if partition is None:
            if not volume.startswith('/'):
                raise ValueError('You should provide a partition name or use partition / path instead.')
            volume = volume.lstrip('/')
            if '/' in volume:
                volume, partition = volume.split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            if not partition:
                raise ValueError('Malformed partition url.')
            return volume.partitions[partition].files.iterate()
        else:
            return volume[partition].objects.iterate()

    def create_volume_directory(self, volume, path=None, project=None, schema=None):
        """
        Create a directory under a file system volume.

        :param str volume: name of the volume.
        :param str path: path of the directory to be created.
        :param str project: project name, if not provided, will be the default project.
        :param str schema: schema name, if not provided, will be the default schema
        :return: directory object.
        """
        from .models import PartedVolume
        if path is None:
            if not volume.startswith('/'):
                raise ValueError('You should provide a valid path.')
            volume = volume.lstrip('/')
            if '/' in volume:
                volume, path = volume.split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            raise ValueError('Only supported under file system volumes.')
        else:
            return volume.create_dir(path)

    def get_volume_file(self, volume, path=None, project=None, schema=None):
        """
        Get a file under a partition of a parted volume, or a file / directory object under a file system volume.

        :param str volume: name of the volume.
        :param str path: path of the directory to be created.
        :param str project: project name, if not provided, will be the default project.
        :param str schema: schema name, if not provided, will be the default schema
        :return: directory object.
        """
        from .models import PartedVolume
        if path is None:
            if not volume.startswith('/'):
                raise ValueError('You should provide a valid path.')
            volume = volume.lstrip('/')
            if '/' in volume:
                volume, path = volume.split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            if '/' not in path:
                raise ValueError('Partition/File format malformed.')
            part, file_name = path.split('/', 1)
            return volume.get_partition(part).files[file_name]
        else:
            return volume[path]

    def move_volume_file(self, old_path, new_path, replication=None, project=None, schema=None):
        """
        Move a file / directory object under a file system volume to another location in the same volume.

        :param str old_path: old path of the volume file.
        :param str new_path: target path of the moved file.
        :param int replication: file replication.
        :param str project: project name, if not provided, will be the default project.
        :param str schema: schema name, if not provided, will be the default schema
        :return: directory object.
        """
        from .models import PartedVolume

        if not new_path.startswith('/'):
            # make relative path absolute
            old_root, _ = old_path.rsplit('/', 1)
            new_path = old_root + '/' + new_path

        if not old_path.startswith('/'):
            raise ValueError('You should provide a valid path.')
        old_volume, old_path = old_path.lstrip('/').split('/', 1)

        new_volume, _ = new_path.lstrip('/').split('/', 1)

        if old_volume != new_volume:
            raise ValueError('Moving between different volumes is not supported.')

        volume = self.get_volume(old_volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            raise ValueError('Only supported under file system volumes.')
        else:
            volume[old_path].move(new_path, replication=replication)

    def delete_volume_file(self, volume, path=None, recursive=False, project=None, schema=None):
        """
        Delete a file / directory object under a file system volume.

        :param str volume: name of the volume.
        :param str path: path of the directory to be created.
        :param bool recursive: if True, recursively delete files
        :param str project: project name, if not provided, will be the default project.
        :param str schema: schema name, if not provided, will be the default schema
        :return: directory object.
        """
        from .models import PartedVolume
        if path is None:
            if not volume.startswith('/'):
                raise ValueError('You should provide a valid path.')
            volume = volume.lstrip('/')
            if '/' in volume:
                volume, path = volume.split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            raise ValueError('Only supported under file system volumes.')
        else:
            volume[path].delete(recursive=recursive)

    def open_volume_reader(self, volume, partition=None, file_name=None, project=None,
                           schema=None, start=None, length=None, **kwargs):
        """
        Open a volume file for read. A file-like object will be returned which can be used to read contents from
        volume files.

        :param str volume: name of the volume
        :param str partition: name of the partition
        :param str file_name: name of the file
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param start: start position
        :param length: length limit
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: CompressOption

        :Example:
        >>> with odps.open_volume_reader('parted_volume', 'partition', 'file') as reader:
        >>>     [print(line) for line in reader]
        """
        from .models import PartedVolume
        if partition is None:
            if not volume.startswith('/'):
                raise ValueError('You should provide a partition name or use partition / path instead.')
            volume = volume.lstrip('/')
            volume, partition = volume.split('/', 1)
            if '/' in partition:
                partition, file_name = partition.rsplit('/', 1)
            else:
                partition, file_name = None, partition
        volume = self.get_volume(volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            if not partition:
                raise ValueError('Malformed partition url.')
            return volume.partitions[partition].open_reader(file_name, start=start, length=length, **kwargs)
        else:
            return volume[partition].open_reader(file_name, start=start, length=length, **kwargs)

    def open_volume_writer(self, volume, partition=None, project=None, schema=None, **kwargs):
        """
        Write data into a volume. This function behaves differently under different types of volumes.

        Under partitioned volumes, all files under a partition should be uploaded in one submission. The method
        returns a writer object with whose `open` method you can open a file inside the volume and write to it,
        or you can use `write` method to write to specific files.

        Under file system volumes, the method returns a file-like object.

        :param str volume: name of the volume
        :param str partition: partition name for partitioned volumes, and path for file system volumes.
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`

        :Example:
        >>> # Writing to partitioned volumes
        >>> with odps.open_volume_writer('parted_volume', 'partition') as writer:
        >>>     # both write methods are acceptable
        >>>     writer.open('file1').write('some content')
        >>>     writer.write('file2', 'some content')
        >>> # Writing to file system volumes
        >>> with odps.open_volume_writer('/fs_volume/dir1/file_name') as writer:
        >>>     writer.write('some content')
        """
        from .models import PartedVolume
        if partition is None:
            if not volume.startswith('/'):
                raise ValueError('You should provide a partition name or use partition / path instead.')
            volume = volume.lstrip('/')
            volume, partition = volume.split('/', 1)
        volume = self.get_volume(volume, project, schema=schema)
        if isinstance(volume, PartedVolume):
            return volume.partitions[partition].open_writer(**kwargs)
        else:
            if '/' in partition:
                partition, file_name = partition.rsplit('/', 1)
            else:
                partition, file_name = None, partition
            return volume[partition].open_writer(file_name, **kwargs)

    def list_xflows(self, project=None, owner=None):
        """
        List xflows of a project which can be filtered by the xflow owner.

        :param str project: project name, if not provided, will be the default project
        :param str owner: Aliyun account
        :return: xflows
        :rtype: list
        """

        project = self.get_project(name=project)
        return project.xflows.iterate(owner=owner)

    def get_xflow(self, name, project=None):
        """
        Get xflow by given name

        :param name: xflow name
        :param project: project name, if not provided, will be the default project
        :return: xflow
        :rtype: :class:`odps.models.XFlow`
        :raise: :class:`odps.errors.NoSuchObject` if not exists

        .. seealso:: :class:`odps.models.XFlow`
        """

        project = self.get_project(name=project)
        return project.xflows[name]

    def exist_xflow(self, name, project=None):
        """
        If the xflow with given name exists or not.

        :param name: xflow name
        :param project: project name, if not provided, will be the default project
        :return: True if exists or False
        :rtype: bool
        """

        project = self.get_project(name=project)
        return name in project.xflows

    def run_xflow(self, xflow_name, xflow_project=None, parameters=None, project=None,
                  hints=None, priority=None):
        """
        Run xflow by given name, xflow project, paremeters asynchronously.

        :param xflow_name: XFlow name
        :type xflow_name: str
        :param xflow_project: the project XFlow deploys
        :type xflow_project: str
        :param parameters: parameters
        :type parameters: dict
        :param project: project name, if not provided, will be the default project
        :param hints: execution hints
        :type hints: dict
        :param priority: instance priority, 9 as default
        :type priority: int
        :return: instance
        :rtype: :class:`odps.models.Instance`

        .. seealso:: :class:`odps.models.Instance`
        """
        project = self.get_project(name=project)
        xflow_project = xflow_project or project
        if isinstance(xflow_project, models.Project):
            xflow_project = xflow_project.name
        return project.xflows.run_xflow(
            xflow_name=xflow_name, xflow_project=xflow_project, project=project, parameters=parameters,
            hints=hints, priority=priority)

    def execute_xflow(self, xflow_name, xflow_project=None, parameters=None, project=None,
                      hints=None, priority=None):
        """
        Run xflow by given name, xflow project, paremeters, block until xflow executed successfully.

        :param xflow_name: XFlow name
        :type xflow_name: str
        :param xflow_project: the project XFlow deploys
        :type xflow_project: str
        :param parameters: parameters
        :type parameters: dict
        :param project: project name, if not provided, will be the default project
        :param hints: execution hints
        :type hints: dict
        :param priority: instance priority, 9 as default
        :type priority: int
        :return: instance
        :rtype: :class:`odps.models.Instance`

        .. seealso:: :class:`odps.models.Instance`
        """

        inst = self.run_xflow(
            xflow_name, xflow_project=xflow_project, parameters=parameters, project=project,
            hints=hints, priority=priority)
        inst.wait_for_success()
        return inst

    def get_xflow_results(self, instance, project=None):
        """
        The result given the results of xflow

        :param instance: instance of xflow
        :type instance: :class:`odps.models.Instance`
        :param project: project name, if not provided, will be the default project
        :return: xflow result
        :rtype: dict
        """

        project = self.get_project(name=project)

        from .models import Instance
        if not isinstance(instance, Instance):
            instance = project.instances[instance]

        return project.xflows.get_xflow_results(instance)

    def get_xflow_sub_instances(self, instance, project=None):
        """
        The result iterates the sub instance of xflow

        :param instance: instance of xflow
        :type instance: :class:`odps.models.Instance`
        :param project: project name, if not provided, will be the default project
        :return: sub instances dictionary
        """
        project = self.get_project(name=project)
        return project.xflows.get_xflow_sub_instances(instance)

    def iter_xflow_sub_instances(self, instance, interval=1, project=None, check=False):
        """
        The result iterates the sub instance of xflow and will wait till instance finish

        :param instance: instance of xflow
        :type instance: :class:`odps.models.Instance`
        :param interval: time interval to check
        :param project: project name, if not provided, will be the default project
        :param bool check: check if the instance is successful
        :return: sub instances dictionary
        """
        project = self.get_project(name=project)
        return project.xflows.iter_xflow_sub_instances(
            instance, interval=interval, check=check
        )

    def delete_xflow(self, name, project=None):
        """
        Delete xflow by given name.

        :param name: xflow name
        :param project: project name, if not provided, will be the default project
        :return: None
        """

        project = self.get_project(name=project)

        return project.xflows.delete(name)

    def list_offline_models(self, project=None, prefix=None, owner=None):
        """
        List offline models of project by optional filter conditions including prefix and owner.

        :param project: project name, if not provided, will be the default project
        :param prefix: prefix of offline model's name
        :param owner: Aliyun account
        :return: offline models
        :rtype: list
        """

        project = self.get_project(name=project)
        return project.offline_models.iterate(name=prefix, owner=owner)

    def get_offline_model(self, name, project=None):
        """
        Get offline model by given name

        :param name: offline model name
        :param project: project name, if not provided, will be the default project
        :return: offline model
        :rtype: :class:`odps.models.ml.OfflineModel`
        :raise: :class:`odps.errors.NoSuchObject` if not exists
        """

        project = self.get_project(name=project)
        return project.offline_models[name]

    def exist_offline_model(self, name, project=None):
        """
        If the offline model with given name exists or not.

        :param name: offline model's name
        :param project: project name, if not provided, will be the default project
        :return: True if offline model exists else False
        :rtype: bool
        """

        project = self.get_project(name=project)
        return name in project.offline_models

    @utils.with_wait_argument
    def copy_offline_model(self, name, new_name, project=None, new_project=None, async_=False):
        """
        Copy current model into a new location.

        :param new_name: name of the new model
        :param new_project: new project name. if absent, original project name will be used
        :param async_: if True, return the copy instance. otherwise return the newly-copied model
        """
        return self.get_offline_model(name, project=project) \
            .copy(new_name, new_project=new_project, async_=async_)

    def delete_offline_model(self, name, project=None, if_exists=False):
        """
        Delete the offline model by given name.

        :param name: offline model's name
        :param if_exists:  will not raise errors when the offline model does not exist, default False
        :param project: project name, if not provided, will be the default project
        :return: None
        """

        project = self.get_project(name=project)
        try:
            return project.offline_models.delete(name)
        except NoSuchObject:
            if not if_exists:
                raise

    def get_logview_host(self):
        """
        Get logview host address.
        :return: logview host address
        """
        if self.endpoint in _logview_host_cache:
            return _logview_host_cache[self.endpoint]

        try:
            logview_host = utils.to_str(
                self.rest.get(self.endpoint + '/logview/host').content
            )
        except:
            logview_host = None
        if not logview_host:
            logview_host = utils.get_default_logview_endpoint(LOGVIEW_HOST_DEFAULT, self.endpoint)
        _logview_host_cache[self.endpoint] = logview_host
        return logview_host

    def get_logview_address(self, instance_id, hours=None, project=None):
        """
        Get logview address by given instance id and hours.

        :param instance_id: instance id
        :param hours:
        :param project: project name, if not provided, will be the default project
        :return: logview address
        :rtype: str
        """
        hours = hours or options.logview_hours
        inst = self.get_instance(instance_id, project=project)
        return inst.get_logview_address(hours=hours)

    def get_project_policy(self, project=None):
        """
        Get policy of a project

        :param project: project name, if not provided, will be the default project
        :return: JSON object
        """
        project = self.get_project(name=project)
        return project.policy

    def set_project_policy(self, policy, project=None):
        """
        Set policy of a project

        :param policy: name of policy.
        :param project: project name, if not provided, will be the default project
        :return: JSON object
        """
        project = self.get_project(name=project)
        project.policy = policy

    def create_role(self, name, project=None):
        """
        Create a role in a project

        :param name: name of the role to create
        :param project: project name, if not provided, will be the default project
        :return: role object created
        """
        project = self.get_project(name=project)
        return project.roles.create(name)

    def list_roles(self, project=None):
        """
        List all roles in a project

        :param project: project name, if not provided, will be the default project
        :return: collection of role objects
        """
        project = self.get_project(name=project)
        return project.roles

    def exist_role(self, name, project=None):
        """
        Check if a role exists in a project

        :param name: name of the role
        :param project: project name, if not provided, will be the default project
        """
        project = self.get_project(name=project)
        return name in project.roles

    def delete_role(self, name, project=None):
        """
        Delete a role in a project

        :param name: name of the role to delete
        :param project: project name, if not provided, will be the default project
        """
        project = self.get_project(name=project)
        project.roles.delete(name)

    def get_role_policy(self, name, project=None):
        """
        Get policy object of a role

        :param name: name of the role
        :param project: project name, if not provided, will be the default project
        :return: JSON object
        """
        project = self.get_project(name=project)
        return project.roles[name].policy

    def set_role_policy(self, name, policy, project=None):
        """
        Get policy object of project

        :param name: name of the role
        :param policy: policy string or JSON object
        :param project: project name, if not provided, will be the default project
        """
        project = self.get_project(name=project)
        project.roles[name].policy = policy

    def list_role_users(self, name, project=None):
        """
        List users who have the specified role.

        :param name: name of the role
        :param project: project name, if not provided, will be the default project
        :return: collection of User objects
        """
        project = self.get_project(name=project)
        return project.roles[name].users

    def create_user(self, name, project=None):
        """
        Add a user into the project

        :param name: user name
        :param project: project name, if not provided, will be the default project
        :return: user created
        """
        project = self.get_project(name=project)
        return project.users.create(name)

    def list_users(self, project=None):
        """
        List users in the project

        :param project: project name, if not provided, will be the default project
        :return: collection of User objects
        """
        project = self.get_project(name=project)
        return project.users

    def exist_user(self, name, project=None):
        """
        Check if a user exists in the project

        :param name: user name
        :param project: project name, if not provided, will be the default project
        """
        project = self.get_project(name=project)
        return name in project.users

    def delete_user(self, name, project=None):
        """
        Delete a user from the project

        :param name: user name
        :param project: project name, if not provided, will be the default project
        """
        project = self.get_project(name=project)
        project.users.delete(name)

    def list_user_roles(self, name, project=None):
        """
        List roles of the specified user

        :param name: user name
        :param project: project name, if not provided, will be the default project
        :return: collection of Role object
        """
        project = self.get_project(name=project)
        return project.users[name].roles

    def get_security_options(self, project=None):
        """
        Get all security options of a project

        :param project: project name, if not provided, will be the default project
        :return: SecurityConfiguration object
        """
        project = self.get_project(name=project)
        return project.security_options

    def get_security_option(self, option_name, project=None):
        """
        Get one security option of a project

        :param option_name: name of the security option. Please refer to ODPS options for more details.
        :param project: project name, if not provided, will be the default project
        :return: option value
        """
        option_name = utils.camel_to_underline(option_name)
        sec_options = self.get_security_options(project=project)
        if not hasattr(sec_options, option_name):
            raise ValueError('Option does not exists.')
        return getattr(sec_options, option_name)

    def set_security_option(self, option_name, value, project=None):
        """
        Set a security option of a project

        :param option_name: name of the security option. Please refer to ODPS options for more details.
        :param value: value of security option to be set.
        :param project: project name, if not provided, will be the default project.
        """
        option_name = utils.camel_to_underline(option_name)
        sec_options = self.get_security_options(project=project)
        if not hasattr(sec_options, option_name):
            raise ValueError('Option does not exists.')
        setattr(sec_options, option_name, value)
        sec_options.update()

    def run_security_query(
        self, query, project=None, schema=None, token=None, hints=None, output_json=True
    ):
        """
        Run a security query to grant / revoke / query privileges. If the query is `install package`
        or `uninstall package`, return a waitable AuthQueryInstance object, otherwise returns
        the result string or json value.

        :param str query: query text
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param bool output_json: parse json for the output
        :return: result string / json object
        """
        project = self.get_project(name=project)
        schema = schema or self.schema
        return project.run_security_query(
            query, schema=schema, token=token, hints=hints, output_json=output_json
        )

    def execute_security_query(
        self, query, project=None, schema=None, token=None, hints=None, output_json=True
    ):
        """
        Execute a security query to grant / revoke / query privileges and returns
        the result string or json value.

        :param str query: query text
        :param str project: project name, if not provided, will be the default project
        :param str schema: schema name, if not provided, will be the default schema
        :param bool output_json: parse json for the output
        :return: result string / json object
        """
        from .models import Project

        instance_or_result = self.run_security_query(
            query, project=project, schema=schema, token=token, hints=hints, output_json=output_json
        )
        if not isinstance(instance_or_result, Project.AuthQueryInstance):
            return instance_or_result
        return instance_or_result.wait_for_success()

    @utils.deprecated(
        'You no longer have to manipulate session instances to use MaxCompute QueryAcceleration. '
        'Try `run_sql_interactive`.'
    )
    def attach_session(self, session_name, taskname=None, hints=None):
        """
        Attach to an existing session.

        :param session_name: The session name.
        :param taskname: The created sqlrt task name. If not provided, the default value is used.
            Mostly doesn't matter, default works.
        :return: A SessionInstance you may execute select tasks within.
        """
        return self._attach_mcqa_session(session_name, task_name=taskname, hints=hints)

    def _attach_mcqa_session(self, session_name=None, task_name=None, hints=None):
        session_name = session_name or models.session.PUBLIC_SESSION_NAME
        task_name = task_name or models.session.DEFAULT_TASK_NAME

        task = models.tasks.SQLRTTask(name=task_name)
        task.update_sql_rt_settings(hints)
        task.update_sql_rt_settings(
            {"odps.sql.session.share.id": session_name, "odps.sql.submit.mode": "script"}
        )
        project = self.get_project()
        return project.instances.create(
            task=task, session_project=project, session_name=session_name
        )

    @utils.deprecated(
        'You no longer have to manipulate session instances to use MaxCompute QueryAcceleration. '
        'Try `run_sql_interactive`.'
    )
    def default_session(self):
        """
        Attach to the default session of your project.

        :return: A SessionInstance you may execute select tasks within.
        """
        return self._get_default_mcqa_session(wait=False)

    def _get_default_mcqa_session(
        self, session_name=None, hints=None, wait=True, service_startup_timeout=60
    ):
        session_name = session_name or models.session.PUBLIC_SESSION_NAME
        if self._default_session is None:
            self._default_session = self._attach_mcqa_session(session_name, hints=hints)
            self._default_session_name = session_name
            if wait:
                self._default_session.wait_for_startup(
                    0.1, service_startup_timeout, max_interval=1
                )
        return self._default_session

    @utils.deprecated(
        'You no longer have to manipulate session instances to use MaxCompute QueryAcceleration. '
        'Try `run_sql_interactive`.'
    )
    def create_session(
        self, session_worker_count, session_worker_memory, session_name=None,
        worker_spare_span=None, taskname=None, hints=None
    ):
        """
        Create session.

        :param session_worker_count: How much workers assigned to the session.
        :param session_worker_memory: How much memory each worker consumes.
        :param session_name: The session name. Not specifying to use its ID as name.
        :param worker_spare_span: format "00-24", allocated workers will be reduced during this time.
            Not specifying to disable this.
        :param taskname: The created sqlrt task name. If not provided, the default value is used.
            Mostly doesn't matter, default works.
        :param hints: Extra hints provided to the session. Parameters of this method will override
            certain hints.
        :return: A SessionInstance you may execute select tasks within.
        """
        return self._create_mcqa_session(
            session_worker_count, session_worker_memory, session_name,
            worker_spare_span, taskname, hints
        )

    def _create_mcqa_session(
        self, session_worker_count, session_worker_memory, session_name=None,
        worker_spare_span=None, task_name=None, hints=None
    ):
        if not task_name:
            task_name = models.session.DEFAULT_TASK_NAME

        session_hints = {
            "odps.sql.session.worker.count": str(session_worker_count),
            "odps.sql.session.worker.memory": str(session_worker_memory),
            "odps.sql.submit.mode": "script",
        }
        if session_name:
            session_hints["odps.sql.session.name"] = session_name
        if worker_spare_span:
            session_hints["odps.sql.session.worker.sparespan"] = worker_spare_span
        task = models.tasks.SQLRTTask(name=task_name)
        task.update_sql_rt_settings(hints)
        task.update_sql_rt_settings(session_hints)
        project = self.get_project()
        return project.instances.create(
            task=task, session_project=project, session_name=session_name
        )

    def run_sql_interactive(self, sql, hints=None, **kwargs):
        """
        Run SQL query in interactive mode (a.k.a MaxCompute QueryAcceleration).
        Won't fallback to offline mode automatically if query not supported or fails
        :param sql: the sql query.
        :param hints: settings for sql query.
        :return: instance.
        """
        cached_is_running = False
        service_name = kwargs.pop('service_name', models.session.PUBLIC_SESSION_NAME)
        task_name = kwargs.pop('task_name', None)
        service_startup_timeout = kwargs.pop('service_startup_timeout', 60)
        force_reattach = kwargs.pop('force_reattach', False)
        if self._default_session != None:
            try:
                cached_is_running = self._default_session.is_running()
            except BaseException:
                pass
        if (
            force_reattach
            or not cached_is_running
            or self._default_session_name != service_name
        ):
            # should reattach, for whatever reason (timed out, terminated, never created,
            # forced using another session)
            self._default_session = self._attach_mcqa_session(service_name, task_name=task_name)
            self._default_session.wait_for_startup(0.1, service_startup_timeout, max_interval=1)
            self._default_session_name = service_name
        return self._default_session.run_sql(sql, hints, **kwargs)

    @utils.deprecated(
        'The method `run_sql_interactive_with_fallback` is deprecated. '
        'Try `execute_sql_interactive` with fallback=True argument instead.'
    )
    def run_sql_interactive_with_fallback(self, sql, hints=None, **kwargs):
        return self.execute_sql_interactive(
            sql, hints=hints, fallback='all', wait_fallback=False, **kwargs
        )

    def execute_sql_interactive(
        self, sql, hints=None, fallback=True, wait_fallback=True, **kwargs
    ):
        """
        Run SQL query in interactive mode (a.k.a MaxCompute QueryAcceleration).
        If query is not supported or fails, and fallback is True,
        will fallback to offline mode automatically

        :param sql: the sql query.
        :param hints: settings for sql query.
        :param fallback: fallback query to non-interactive mode, True by default.
            Both boolean type and policy names separated by commas are acceptable.
        :param bool wait_fallback: wait fallback instance to finish, True by default.
        :return: instance.
        """
        from .models.session import FallbackMode, FallbackPolicy

        if isinstance(fallback, (six.string_types, set, list, tuple)):
            fallback_policy = FallbackPolicy(fallback)
        elif fallback is False:
            fallback_policy = None
        elif fallback is True:
            fallback_policy = FallbackPolicy("all")
        else:
            assert isinstance(fallback, FallbackPolicy)
            fallback_policy = fallback

        inst = None
        use_tunnel = kwargs.pop("tunnel", True)
        fallback_callback = kwargs.pop("fallback_callback", None)
        offline_hints = kwargs.pop("offline_hints", None) or {}
        try:
            inst = self.run_sql_interactive(sql, hints=hints, **kwargs)
            inst.wait_for_success(interval=0.1, max_interval=1)
            try:
                rd = inst.open_reader(tunnel=use_tunnel, limit=True)
                if not rd:
                    raise ODPSError('Get sql result fail')
            except errors.InstanceTypeNotSupported:
                # sql is not a select, just skip creating reader
                pass
            return inst
        except BaseException as ex:
            if fallback_policy is None:
                raise
            fallback_mode = fallback_policy.get_mode_from_exception(ex)
            if fallback_mode is None:
                raise
            elif fallback_mode == FallbackMode.INTERACTIVE:
                kwargs["force_reattach"] = True
                return self.execute_sql_interactive(
                    sql, hints=hints, fallback=fallback, wait_fallback=wait_fallback, **kwargs
                )
            else:
                kwargs.pop("service_name", None)
                kwargs.pop("force_reattach", None)
                kwargs.pop("service_startup_timeout", None)
                hints = copy.copy(offline_hints or hints or {})
                hints["odps.task.sql.sqa.enable"] = "false"

                if fallback_callback is not None:
                    fallback_callback(inst, ex)

                if inst is not None:
                    hints["odps.sql.session.fallback.instance"] = "%s_%s" % (inst.id, inst.subquery_id)
                else:
                    hints["odps.sql.session.fallback.instance"] = "fallback4AttachFailed"
                inst = self.execute_sql(sql, hints=hints, **kwargs)
                if wait_fallback:
                    inst.wait_for_success()
                return inst

    @classmethod
    def _build_account(cls, access_id, secret_access_key):
        return accounts.AliyunAccount(access_id, secret_access_key)

    def to_global(self, overwritable=False):
        options.is_global_account_overwritable = overwritable
        options.account = self.account
        options.default_project = self.project
        # use _schema to avoid requesting for tenant options
        options.default_schema = self._schema
        options.endpoint = self.endpoint
        options.logview_host = self.logview_host
        options.tunnel.endpoint = self._tunnel_endpoint
        options.app_account = self.app_account
        options.region_name = self.region_name

    @classmethod
    def from_global(cls):
        if options.account is not None and options.default_project is not None:
            return cls._from_account(
                options.account,
                options.default_project,
                endpoint=options.endpoint,
                schema=options.default_schema,
                tunnel_endpoint=options.tunnel.endpoint,
                logview_host=options.logview_host,
                app_account=options.app_account,
                region_name=options.region_name,
            )
        else:
            return None

    @classmethod
    def from_environments(cls):
        try:
            account = accounts.BearerTokenAccount.from_environments()
            if not account:
                raise KeyError('ODPS_BEARER_TOKEN')
            project = os.getenv('ODPS_PROJECT_NAME')
            endpoint = os.environ['ODPS_ENDPOINT']
            tunnel_endpoint = os.getenv('ODPS_TUNNEL_ENDPOINT')
            return cls(None, None, account=account, project=project,
                       endpoint=endpoint, tunnel_endpoint=tunnel_endpoint)
        except KeyError:
            return None


def _get_odps_from_model(self):
    cur = self
    while cur is not None and not isinstance(cur, models.Project):
        cur = cur.parent
    return cur.odps if cur else None


models.RestModel.odps = property(fget=_get_odps_from_model)
del _get_odps_from_model

try:
    from .internal.core import *  # noqa: F401
except ImportError:
    pass

try:
    from .mars_extension import create_mars_cluster
    setattr(ODPS, 'create_mars_cluster', create_mars_cluster)
except ImportError:
    pass
try:
    from .mars_extension import persist_mars_dataframe
    setattr(ODPS, 'persist_mars_dataframe', persist_mars_dataframe)
except ImportError:
    pass
try:
    from .mars_extension import to_mars_dataframe
    setattr(ODPS, 'to_mars_dataframe', to_mars_dataframe)
except ImportError:
    pass
try:
    from .mars_extension import run_script_in_mars
    setattr(ODPS, 'run_script_in_mars', run_script_in_mars)
except ImportError:
    pass
try:
    from .mars_extension import run_mars_job
    setattr(ODPS, 'run_mars_job', run_mars_job)
except ImportError:
    pass
try:
    from .mars_extension import list_mars_instances
    setattr(ODPS, 'list_mars_instances', list_mars_instances)
except ImportError:
    pass
try:
    from .mars_extension import sql_to_mars_dataframe
    setattr(ODPS, 'sql_to_mars_dataframe', sql_to_mars_dataframe)
except ImportError:
    pass


DEFAULT_ENDPOINT = os.getenv('ODPS_ENDPOINT', os.getenv('PYODPS_ENDPOINT', DEFAULT_ENDPOINT))



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os

from .. import compat
from ..compat import six
from ..config import options
from .resource import FileResource

RESOURCE_SIZE_MAX = 512 * 1024 * 1024  # a single resource's size must be at most 512M


class ResourceFile(object):
    __slots__ = (
        'resource', 'mode', '_opened', 'size', '_open_binary', '_encoding', '_overwrite'
    )

    def __init__(self, resource, mode='r', encoding='utf-8', overwrite=None):
        self.resource = resource

        self._open_binary = 'b' in mode
        mode = mode.replace('b', '')
        self.mode = FileResource.Mode(mode)
        self._encoding = encoding

        if self.mode in (FileResource.Mode.WRITE, FileResource.Mode.TRUNCEREADWRITE):
            self.size = 0
        else:
            self.resource._reload_size()
        self._opened = True
        self._overwrite = overwrite

    def _convert(self, content):
        if self._open_binary and isinstance(content, six.text_type):
            return content.encode(self._encoding)
        elif not self._open_binary and isinstance(content, six.binary_type):
            return content.decode(self._encoding)
        return content

    def _read_resource(self, offset=None, read_size=None):
        return self.resource.parent.read_resource(
            self.resource,
            text_mode=not self._open_binary,
            encoding=self._encoding,
            offset=offset,
            read_size=read_size,
        )

    def _new_buffer(self, content=None):
        io_clz = six.BytesIO if self._open_binary else six.StringIO
        return io_clz() if content is None else io_clz(content)

    def read(self, size=-1):
        raise NotImplementedError

    def readline(self, size=-1):
        raise NotImplementedError

    def readlines(self, sizehint=-1):
        raise NotImplementedError

    def write(self, content):
        raise NotImplementedError

    def writelines(self, seq):
        raise NotImplementedError

    def seek(self, pos, whence=compat.SEEK_SET):  # io.SEEK_SET
        raise NotImplementedError

    def seekable(self):
        raise NotImplementedError

    def tell(self):
        raise NotImplementedError

    def truncate(self, size=None):
        raise NotImplementedError

    def flush(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def _iter(self):
        raise NotImplementedError

    def __iter__(self):
        return self._iter()

    def _next(self):
        raise NotImplementedError

    def __next__(self):
        return self._next()

    next = __next__

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


class LocalResourceFile(ResourceFile):
    __slots__ = "_fp", "_need_commit"

    def __init__(self, resource, mode='r', encoding='utf-8', overwrite=None):
        super(LocalResourceFile, self).__init__(
            resource, mode=mode, encoding=encoding, overwrite=overwrite
        )

        if self.mode in (FileResource.Mode.WRITE, FileResource.Mode.TRUNCEREADWRITE):
            self._fp = self._new_buffer()
        else:
            self._fp = self._read_resource()
            self._sync_size()

        self._need_commit = False

    def _sync_size(self):
        curr_pos = self.tell()
        self.seek(0, compat.SEEK_END)
        self.size = self.tell()
        self.seek(curr_pos)

    def read(self, size=-1):
        return self._fp.read(size)

    def readline(self, size=-1):
        return self._fp.readline(size)

    def readlines(self, sizehint=-1):
        return self._fp.readlines(sizehint)

    def _check_size(self):
        if self.size > RESOURCE_SIZE_MAX:
            raise IOError(
                "Single resource's max size is %sM" % (RESOURCE_SIZE_MAX / (1024 ** 2))
            )

    def write(self, content):
        content = self._convert(content)

        length = len(content)
        if self.mode in (FileResource.Mode.APPEND, FileResource.Mode.APPENDREADWRITE):
            self.seek(0, compat.SEEK_END)

        if length > 0:
            self._need_commit = True

        res = self._fp.write(content)
        self._sync_size()
        self._check_size()
        return res

    def writelines(self, seq):
        seq = [self._convert(s) for s in seq]

        length = sum(len(s) for s in seq)
        if self.mode in (FileResource.Mode.APPEND, FileResource.Mode.APPENDREADWRITE):
            self.seek(0, compat.SEEK_END)

        if length > 0:
            self._need_commit = True

        res = self._fp.writelines(seq)
        self._sync_size()
        self._check_size()
        return res

    def seek(self, pos, whence=compat.SEEK_SET):
        return self._fp.seek(pos, whence)

    def seekable(self):
        return True

    def tell(self):
        return self._fp.tell()

    def truncate(self, size=None):
        curr_pos = self.tell()
        self._fp.truncate(size)

        self.seek(0, compat.SEEK_END)
        self.size = self.tell()

        self.seek(curr_pos)

        self._need_commit = True

    def flush(self):
        if self._need_commit:
            is_create = self.resource._is_create()

            resources = self.resource.parent

            if is_create:
                resources.create(obj=self.resource, file_obj=self._fp)
            else:
                resources.update(obj=self.resource, file_obj=self._fp)

            self._need_commit = False

    def close(self):
        if not self._opened:
            # already closed
            return

        self.flush()

        self._fp = None
        self.size = 0
        self._need_commit = False
        self._opened = False

    def _iter(self):
        return self._fp.__iter__()

    def _next(self):
        return next(self._fp)


class StreamResourceFile(ResourceFile):
    __slots__ = (
        "_md5_digest", "_buffer", "_buffered_size", "_resource_parts",
        "_resource_counter", "_chunk_size", "_is_source_exhausted",
        "_source_offset",
    )

    def __init__(self, resource, mode='r', encoding='utf-8', overwrite=None):
        mode = mode.replace("+", "")

        super(StreamResourceFile, self).__init__(
            resource, mode=mode, encoding=encoding, overwrite=overwrite
        )

        if self.mode not in (FileResource.Mode.READ, FileResource.Mode.WRITE):
            raise compat.UnsupportedOperation(
                "Unsupported access mode %s under streaming mode" % mode
            )

        self._md5_digest = hashlib.md5()
        self._resource_parts = []
        self._resource_counter = 0
        self._buffered_size = 0
        self._chunk_size = options.resource_chunk_size
        self._rebuild_buffer()

        self._is_source_exhausted = False
        self._source_offset = 0

    def _rebuild_buffer(self):
        self._buffer, buffer = self._new_buffer(), getattr(self, "_buffer", None)
        self._buffered_size = 0
        return buffer

    def _build_part_resource_name(self):
        name = "%s.part.tmp.%06d" % (self.resource.name, self._resource_counter)
        self._resource_counter += 1
        return name

    def _load_next_offset(self):
        if self._is_source_exhausted:
            return

        buf = self.resource.parent.read_resource(
            self.resource, offset=self._source_offset, read_size=self._chunk_size
        )
        self._rebuild_buffer()
        self._buffer.write(self._convert(buf.read()))
        self._buffer.seek(0, os.SEEK_SET)
        self._buffered_size = buf.tell()
        self._source_offset += buf.tell()
        self._is_source_exhausted = buf.is_eof

    def read(self, size=-1):
        buf = self._new_buffer()
        size_to_read = size
        while size_to_read != 0:
            if self._buffered_size == 0:
                self._load_next_offset()
            if self._buffered_size == 0:
                break
            res = self._buffer.read(size_to_read)

            buf.write(res)

            res_len = len(res)
            if size_to_read > 0:
                size_to_read -= res_len
            self._buffered_size -= res_len
        return buf.getvalue()

    def _is_line_terminated(self, line):
        terminator = b'\n' if self._open_binary else os.linesep
        return line.endswith(terminator)

    def readline(self, size=-1):
        line_buf = self._new_buffer()
        size_to_read = size
        while size_to_read != 0:
            if self._buffered_size == 0:
                self._load_next_offset()
            if self._buffered_size == 0:
                break

            res = self._buffer.readline(size_to_read)
            # read and concatenate to existing line
            line_buf.write(res)

            res_len = len(res)
            if size_to_read > 0:
                size_to_read -= res_len
            self._buffered_size -= res_len

            if self._is_line_terminated(res):
                break
        return line_buf.getvalue()

    def readlines(self, sizehint=-1):
        if sizehint == 0:
            return []

        lines_buf = []
        lines_to_read = sizehint
        while lines_to_read != 0:
            if self._buffered_size == 0:
                self._load_next_offset()
            if self._buffered_size == 0:
                break

            old_pos = self._buffer.tell()
            lines = self._buffer.readlines(lines_to_read)
            if not lines:
                break

            self._buffered_size -= self._buffer.tell() - old_pos

            if self._is_line_terminated(lines[-1]):
                lines_to_read -= len(lines)
                has_terminator = True
            else:
                # last line not complete
                lines_to_read -= len(lines) - 1
                has_terminator = False
            lines_to_read = max(-1, lines_to_read)
            lines_buf.append((lines, has_terminator))

        # merge fraction of lines
        last_has_terminator = True
        res_line_groups = []
        for lines, has_terminator in lines_buf:
            if not last_has_terminator:
                res_line_groups[-1].append(lines[0])
                lines = lines[1:]
            if lines:
                res_line_groups.extend([l] for l in lines)
            last_has_terminator = has_terminator

        res_lines = []
        sep = b"" if self._open_binary else ""
        for lg in res_line_groups:
            res_lines.append(sep.join(lg))

        return res_lines

    def write(self, content):
        if self._buffered_size >= self._chunk_size:
            self.flush()

        content = self._convert(content)

        last_pos = self._buffer.tell()
        self._buffer.write(content)
        content_size = self._buffer.tell() - last_pos

        self._buffered_size += content_size
        self.size += content_size

    def writelines(self, seq):
        if self._buffered_size >= self._chunk_size:
            self.flush()

        seq = [self._convert(s) for s in seq]

        last_pos = self._buffer.tell()
        self._buffer.writelines(seq)
        content_size = self._buffer.tell() - last_pos

        self._buffered_size += content_size
        self.size += content_size

    def seek(self, pos, whence=compat.SEEK_SET):
        raise compat.UnsupportedOperation("File or stream is not seekable.")

    def seekable(self):
        return False

    def tell(self):
        return self.size

    def truncate(self, size=None):
        raise compat.UnsupportedOperation("File or stream is not seekable.")

    def flush(self):
        value = self._rebuild_buffer()

        if value.tell() > 0:
            res = self.resource.parent.create(
                name=self._build_part_resource_name(), type="file", temp=True, part=True, fileobj=value
            )
            self._resource_parts.append(res)
            if self._open_binary:
                self._md5_digest.update(value.getvalue())
            else:
                self._md5_digest.update(value.getvalue().encode(self._encoding))

    def close(self):
        if self.mode == FileResource.Mode.READ or not self._opened:
            return

        self.flush()
        if self._overwrite is None:
            self._overwrite = not self.resource._is_create()

        self.resource.parent.merge_part_files(
            self.resource,
            self._resource_parts,
            self._md5_digest.hexdigest(),
            overwrite=self._overwrite,
        )
        self._opened = False

    def _iter(self):
        return self

    def _next(self):
        line = self.readline()
        if not line:
            raise StopIteration
        return line



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cpython.datetime cimport datetime
from libc.stdint cimport int64_t
from libc.time cimport tm


cdef class CMillisecondsConverter:
    cdef:
        object _local_tz, _tz
        bint _use_default_tz
        bint _default_tz_local
        bint _allow_antique
        bint _is_dst
        bint _tz_has_localize

    cdef int _build_tm_struct(self, datetime dt, tm *p_tm) except? -1
    cpdef int64_t to_milliseconds(self, datetime dt) except? -1
    cpdef datetime from_milliseconds(self, int64_t milliseconds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tokenize
import ast
import operator
from collections import OrderedDict

from . import element as ele
from .expressions import Expr, ExpressionError
from ...compat import reduce, StringIO

LOCAL_TAG = '_local_var_'


def _tokenize_str(reader):
    for toknum, tokval, _, _, _ in tokenize.generate_tokens(reader):
        if toknum == tokenize.OP:
            if tokval == '@':
                tokval = LOCAL_TAG
            if tokval == '&':
                toknum = tokenize.NAME
                tokval = 'and'
            elif tokval == '|':
                toknum = tokenize.NAME
                tokval = 'or'
        yield toknum, '==' if tokval == '=' else tokval


class ExprVisitor(ast.NodeVisitor):
    _op_handlers = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.FloorDiv: operator.floordiv,
        ast.mod: operator.mod,
        ast.Pow: operator.pow,

        ast.Eq: operator.eq,
        ast.NotEq: operator.ne,
        ast.Lt: operator.lt,
        ast.LtE: operator.le,
        ast.Gt: operator.gt,
        ast.GtE: operator.ge,
        ast.In: ele._isin,
        ast.NotIn: ele._notin,

        ast.UAdd: operator.pos,
        ast.USub: operator.neg,
        ast.Invert: operator.invert,

        ast.And: operator.and_,
        ast.Or: operator.or_
    }

    def __init__(self, env):
        self.env = env

    def _preparse(self, expr):
        reader = StringIO(expr).readline
        return tokenize.untokenize(list(_tokenize_str(reader)))

    def eval(self, expr, rewrite=True):
        if rewrite:
            expr = self._preparse(expr)
        node = ast.fix_missing_locations(ast.parse(expr))
        return self.visit(node)

    def get_named_object(self, obj_name):
        raise NotImplementedError

    def visit(self, node):
        if isinstance(node, Expr):
            return node
        node_name = node.__class__.__name__
        method = 'visit_' + node_name
        try:
            visitor = getattr(self, method)
        except AttributeError:
            raise ExpressionError('Query string contains unsupported syntax: {}'.format(node_name))
        return visitor(node)

    def visit_Module(self, node):
        if len(node.body) != 1:
            raise SyntaxError('Only a single expression is allowed')
        expr = node.body[0]
        return self.visit(expr)

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return self._op_handlers[type(node.op)](left, right)

    def visit_Call(self, node):
        from .expressions import ReprWrapper
        func = self.visit(node.func)
        if not isinstance(func, ReprWrapper):
            raise SyntaxError('Calling none-expression methods is not allowed.')
        args = [self.visit(n) for n in node.args]
        kwargs = OrderedDict([(kw.arg, self.visit(kw.value)) for kw in node.keywords])
        return func(*args, **kwargs)

    def visit_Compare(self, node):
        ops = node.ops
        comps = node.comparators

        if len(comps) == 1:
            binop = ast.BinOp(op=ops[0], left=node.left, right=comps[0])
            return self.visit(binop)

        left = node.left
        values = []
        for op, comp in zip(ops, comps):
            new_node = ast.Compare(comparators=[comp], left=left, ops=[op])
            left = comp
            values.append(new_node)
        return self.visit(ast.BoolOp(op=ast.And(), values=values))

    def visit_BoolOp(self, node):
        def func(lhs, rhs):
            binop = ast.BinOp(op=node.op, left=lhs, right=rhs)
            return self.visit(binop)
        return reduce(func, node.values)

    def visit_UnaryOp(self, node):
        oprand = self.visit(node.operand)
        return self._op_handlers[type(node.op)](oprand)

    def visit_Name(self, node):
        if node.id.startswith(LOCAL_TAG):
            local_name = node.id.replace(LOCAL_TAG, '')
            return self.env[local_name]
        if node.id in ['True', 'False', 'None']:
            return eval(node.id)
        return self.get_named_object(node.id)

    def visit_NameConstant(self, node):
        return node.value

    def visit_Num(self, node):
        return node.n

    def visit_Str(self, node):
        return node.s

    def visit_Constant(self, node):
        return node.value

    def visit_List(self, node):
        return [self.visit(e) for e in node.elts]

    visit_Tuple = visit_List

    def visit_Attribute(self, node):
        attr = node.attr
        value = node.value

        ctx = node.ctx
        if isinstance(ctx, ast.Load):
            resolved = self.visit(value)
            return getattr(resolved, attr)

        raise ValueError("Invalid Attribute context {0}".format(ctx.__name__))

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        sub = self.visit(node.slice)
        return value[sub]

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Slice(self, node):
        lower = node.lower
        if lower is not None:
            lower = self.visit(lower)
        upper = node.upper
        if upper is not None:
            upper = self.visit(upper)
        step = node.step
        if step is not None:
            step = self.visit(step)

        return slice(lower, upper, step)


class CollectionVisitor(ExprVisitor):
    def __init__(self, collection, env):
        super(CollectionVisitor, self).__init__(env)
        self.collection = collection

    def get_named_object(self, obj_name):
        return self.collection._get_field(obj_name)


class SequenceVisitor(ExprVisitor):
    def __init__(self, sequence, env):
        super(SequenceVisitor, self).__init__(env)
        self.sequence = sequence

    def get_named_object(self, obj_name):
        return getattr(self.sequence, obj_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct

from ..crc import Crc32c, Crc32
from .. import utils


class Checksum(object):
    TRUE = bytearray([1])
    FALSE = bytearray([0])
    
    def __init__(self, method='crc32c'):
        self.crc = Crc32c() if method.lower() == 'crc32c' else Crc32()

    def _mode(self):
        # use for UT to check if use c extension
        try:
            from ..src.crc32c_c import Crc32c

            return 'c' if isinstance(self.crc, Crc32c) else 'py'
        except ImportError:
            return 'py'

    def update_bool(self, val):
        assert isinstance(val, bool)

        val = self.TRUE if val else self.FALSE
        self._update(val)

    def update_int(self, val):
        val = struct.pack('<i', val)
        self._update(val)

    def update_long(self, val):
        val = struct.pack('<q', val)
        self._update(val)

    def update_float(self, val):
        val = struct.pack('<f', val)
        self._update(val)

    def update_double(self, val):
        val = struct.pack('<d', val)
        self._update(val)

    def _update(self, b):
        # update crc without type checking
        self.crc.update(bytearray(b))
            
    def update(self, b):
        b = utils.to_binary(b)
        self._update(b)
        
    def getvalue(self):
        return self.crc.getvalue()
    
    def reset(self):
        return self.crc.reset()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from .. import serializers
from ..compat import Enum


class StorageTier(Enum):
    STANDARD = "standard"
    LOWFREQENCY = "lowfrequency"
    LONGTERM = "longterm"


class StorageTierInfo(serializers.JSONSerializableModel):
    __slots__ = "storage_size",

    _storage_tier_size_names = {
        StorageTier.STANDARD: "StandardSize",
        StorageTier.LOWFREQENCY: "LowFrequencySize",
        StorageTier.LONGTERM: "LongTermSize",
    }

    _storage_tier_charge_size_names = {
        StorageTier.STANDARD: "chargableAliveDataSize",
        StorageTier.LOWFREQENCY: "chargableLowFreqStorageSize",
        StorageTier.LONGTERM: "chargableLongTermStorageSize",
    }

    storage_tier = serializers.JSONNodeField(
        'StorageTier', parse_callback=lambda x: StorageTier(x.lower()) if x else None
    )
    last_modified_time = serializers.JSONNodeField(
        'StorageLastModifiedTime',
        parse_callback=lambda x: datetime.datetime.fromtimestamp(int(x)),
    )

    def __init__(self, **kwargs):
        super(StorageTierInfo, self).__init__(**kwargs)
        self.storage_size = dict()

    @classmethod
    def deserial(cls, content, obj=None, **kw):
        res = super(StorageTierInfo, cls).deserial(content, obj=obj, **kw)
        for key_defs in (cls._storage_tier_size_names, cls._storage_tier_charge_size_names):
            for tier, key in key_defs.items():
                if key not in content:
                    continue
                res.storage_size[tier] = int(content[key])
        return res if res.storage_tier is not None or res.storage_size else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import numpy as np
import pandas as pd

from mars.utils import parse_readable_size, ceildiv
from mars.serialize import (
    StringField,
    Int64Field,
    SeriesField,
    DictField,
    BoolField,
    ListField,
    ValueType,
)
from mars.dataframe import ArrowStringDtype
from mars.dataframe.operands import DataFrameOperandMixin, DataFrameOperand
from mars.dataframe.utils import (
    parse_index,
    standardize_range_index,
    arrow_table_to_pandas_dataframe,
)
from mars.optimizes.runtime.dataframe import DataSourceHeadRule

try:
    from mars.dataframe.datasource.core import (
        HeadOptimizedDataSource,
        ColumnPruneSupportedDataSourceMixin,
    )

    BASES = (HeadOptimizedDataSource, ColumnPruneSupportedDataSourceMixin)
    head_can_be_opt = True
except ImportError:
    BASES = (DataFrameOperand, DataFrameOperandMixin)
    head_can_be_opt = False
try:
    from mars.core import OutputType

    _output_type_kw = dict(_output_types=[OutputType.dataframe])
except ImportError:
    from mars.dataframe.operands import ObjectType

    _output_type_kw = dict(_object_type=ObjectType.dataframe)

from ....df.backends.odpssql.types import odps_type_to_df_type
from ....errors import ODPSError
from ....utils import to_str, to_timestamp
from ...utils import filter_partitions, check_partition_exist

logger = logging.getLogger(__name__)

READ_CHUNK_LIMIT = 64 * 1024**2
MAX_CHUNK_SIZE = 512 * 1024**2

ORC_COMPRESSION_RATIO = 5
STRING_FIELD_OVERHEAD = 50

_Base = type("_DataSource", BASES, dict())


class DataFrameReadTable(_Base):
    _op_type_ = 123450

    _odps_params = DictField("odps_params")
    _table_name = StringField("table_name")
    _partition_spec = StringField("partition_spec")
    _dtypes = SeriesField("dtypes")
    _add_offset = BoolField("add_offset")
    _columns = ListField("columns")
    _nrows = Int64Field("nrows")
    _use_arrow_dtype = BoolField("use_arrow_dtype")
    _string_as_binary = BoolField("string_as_binary")
    _append_partitions = BoolField("append_partitions")
    _last_modified_time = Int64Field("last_modified_time")
    _with_split_meta_on_tile = BoolField("with_split_meta_on_tile")
    _index_columns = ListField("index_columns", ValueType.string)
    _index_dtypes = SeriesField("index_dtypes")

    def __init__(
        self,
        odps_params=None,
        table_name=None,
        partition_spec=None,
        columns=None,
        dtypes=None,
        nrows=None,
        sparse=None,
        add_offset=True,
        use_arrow_dtype=None,
        string_as_binary=None,
        memory_scale=None,
        append_partitions=None,
        last_modified_time=None,
        with_split_meta_on_tile=False,
        index_columns=None,
        index_dtypes=None,
        **kw
    ):
        kw.update(_output_type_kw)
        super(DataFrameReadTable, self).__init__(
            _odps_params=odps_params,
            _table_name=table_name,
            _partition_spec=partition_spec,
            _columns=columns,
            _dtypes=dtypes,
            _nrows=nrows,
            _sparse=sparse,
            _use_arrow_dtype=use_arrow_dtype,
            _string_as_binary=string_as_binary,
            _add_offset=add_offset,
            _append_partitions=append_partitions,
            _last_modified_time=last_modified_time,
            _memory_scale=memory_scale,
            _with_split_meta_on_tile=with_split_meta_on_tile,
            _index_columns=index_columns,
            _index_dtypes=index_dtypes,
            **kw
        )

    @property
    def retryable(self):
        return False

    @property
    def odps_params(self):
        return self._odps_params

    @property
    def table_name(self):
        return self._table_name

    @property
    def partition(self):
        return getattr(self, "_partition_spec", None)

    @property
    def columns(self):
        return self._columns

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def nrows(self):
        return self._nrows

    @property
    def use_arrow_dtype(self):
        return self._use_arrow_dtype

    @property
    def string_as_binary(self):
        return self._string_as_binary

    @property
    def add_offset(self):
        return self._add_offset

    @property
    def append_partitions(self):
        return self._append_partitions

    @property
    def with_split_meta_on_tile(self):
        return self._with_split_meta_on_tile

    @property
    def index_columns(self):
        return self._index_columns

    @property
    def index_dtypes(self):
        return self._index_dtypes

    def get_columns(self):
        return self._columns

    def set_pruned_columns(self, columns):
        self._columns = columns

    def __call__(self, shape, chunk_bytes=None, chunk_size=None):
        import numpy as np
        import pandas as pd

        if not self.index_columns:
            if np.isnan(shape[0]):
                index_value = parse_index(pd.RangeIndex(0))
            else:
                index_value = parse_index(pd.RangeIndex(shape[0]))
        elif len(self.index_columns) == 1:
            index_value = parse_index(pd.Index([]).astype(self.index_dtypes[0]))
        else:
            idx = pd.MultiIndex.from_frame(
                pd.DataFrame([], columns=self.index_columns).astype(self.index_dtypes)
            )
            index_value = parse_index(idx)

        columns_value = parse_index(self.dtypes.index, store_data=True)
        return self.new_dataframe(
            None,
            shape,
            dtypes=self.dtypes,
            index_value=index_value,
            columns_value=columns_value,
            chunk_bytes=chunk_bytes,
            chunk_size=chunk_size,
        )

    @classmethod
    def _tile_cupid(cls, op):
        from odps import ODPS
        from odps.accounts import BearerTokenAccount
        from cupid import CupidSession, context
        from cupid.errors import CupidError
        from mars.context import get_context

        cupid_ctx = context()

        bearer_token = cupid_ctx.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            None,
            None,
            account=account,
            project=odps_params["project"],
            endpoint=endpoint,
        )
        cupid_session = CupidSession(o)

        mars_context = get_context()

        df = op.outputs[0]
        split_size = df.extra_params.chunk_bytes or READ_CHUNK_LIMIT

        out_dtypes = df.dtypes
        out_shape = df.shape
        out_columns_value = df.columns_value
        if op.columns is not None:
            out_dtypes = out_dtypes[op.columns]
            out_shape = (df.shape[0], len(op.columns))
            out_columns_value = parse_index(out_dtypes.index, store_data=True)

        table_obj = o.get_table(op.table_name)
        if not table_obj.table_schema.partitions:
            data_srcs = [table_obj]
        elif op.partition is not None and check_partition_exist(
            table_obj, op.partition
        ):
            data_srcs = [table_obj.get_partition(op.partition)]
        else:
            data_srcs = list(table_obj.partitions)
            if op.partition is not None:
                data_srcs = filter_partitions(o, data_srcs, op.partition)

        out_chunks = []
        chunk_idx = 0

        for data_src in data_srcs:
            try:
                data_store_size = data_src.size
            except ODPSError:
                # fail to get data size, just ignore
                pass
            else:
                if data_store_size < split_size and mars_context is not None:
                    # get worker counts
                    worker_count = max(len(mars_context.get_worker_addresses()), 1)
                    # data is too small, split as many as number of cores
                    split_size = data_store_size // worker_count
                    # at least 1M
                    split_size = max(split_size, 1 * 1024**2)
                    logger.debug(
                        "Input data size is too small, split_size is %s", split_size
                    )

            logger.debug(
                "Start creating download session of table %s from cupid, "
                "columns: %s",
                op.table_name,
                op.columns,
            )
            while True:
                try:
                    download_session = cupid_session.create_download_session(
                        data_src,
                        split_size=split_size,
                        columns=op.columns,
                        with_split_meta=op.with_split_meta_on_tile,
                    )
                    break
                except CupidError:
                    logger.debug(
                        "The number of splits exceeds 100000, split_size is %s",
                        split_size,
                    )
                    if split_size >= MAX_CHUNK_SIZE:
                        raise
                    else:
                        split_size *= 2

            logger.debug(
                "%s table splits have been created.", str(len(download_session.splits))
            )

            meta_chunk_rows = [
                split.meta_row_count for split in download_session.splits
            ]
            if np.isnan(out_shape[0]):
                est_chunk_rows = meta_chunk_rows
            else:
                sp_file_sizes = np.array(
                    [
                        sp.split_file_end - sp.split_file_start
                        for sp in download_session.splits
                    ]
                )
                total_size = sp_file_sizes.sum()
                ratio_chunk_rows = (sp_file_sizes * out_shape[0] // total_size).tolist()
                est_chunk_rows = [
                    mr if mr is not None else rr
                    for mr, rr in zip(meta_chunk_rows, ratio_chunk_rows)
                ]

            partition_spec = (
                str(data_src.partition_spec)
                if getattr(data_src, "partition_spec", None)
                else None
            )

            logger.warning("Estimated chunk rows: %r", est_chunk_rows)

            if len(download_session.splits) == 0:
                logger.debug("Table %s has no data", op.table_name)
                chunk_op = DataFrameReadTableSplit()
                index_value = parse_index(pd.RangeIndex(0))
                columns_value = parse_index(out_dtypes.index, store_data=True)
                out_chunk = chunk_op.new_chunk(
                    None,
                    shape=(np.nan, out_shape[1]),
                    dtypes=op.dtypes,
                    index_value=index_value,
                    columns_value=columns_value,
                    index=(chunk_idx, 0),
                )
                out_chunks.append(out_chunk)
                chunk_idx += 1
            else:
                for idx, split in enumerate(download_session.splits):
                    chunk_op = DataFrameReadTableSplit(
                        cupid_handle=to_str(split.handle),
                        split_index=split.split_index,
                        split_file_start=split.split_file_start,
                        split_file_end=split.split_file_end,
                        schema_file_start=split.schema_file_start,
                        schema_file_end=split.schema_file_end,
                        add_offset=op.add_offset,
                        dtypes=out_dtypes,
                        sparse=op.sparse,
                        split_size=split_size,
                        string_as_binary=op.string_as_binary,
                        use_arrow_dtype=op.use_arrow_dtype,
                        estimate_rows=est_chunk_rows[idx],
                        partition_spec=partition_spec,
                        append_partitions=op.append_partitions,
                        meta_raw_size=split.meta_raw_size,
                        nrows=meta_chunk_rows[idx] or op.nrows,
                        memory_scale=op.memory_scale,
                        index_columns=op.index_columns,
                    )
                    # the chunk shape is unknown
                    index_value = parse_index(pd.RangeIndex(0))
                    columns_value = parse_index(out_dtypes.index, store_data=True)
                    out_chunk = chunk_op.new_chunk(
                        None,
                        shape=(np.nan, out_shape[1]),
                        dtypes=out_dtypes,
                        index_value=index_value,
                        columns_value=columns_value,
                        index=(chunk_idx, 0),
                    )
                    chunk_idx += 1
                    out_chunks.append(out_chunk)

        if op.add_offset:
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_shape[1],))
        return new_op.new_dataframes(
            None,
            shape=out_shape,
            dtypes=op.dtypes,
            index_value=df.index_value,
            columns_value=out_columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_tunnel(cls, op):
        from odps import ODPS

        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            odps_params["access_id"],
            odps_params["secret_access_key"],
            project=odps_params["project"],
            endpoint=endpoint,
        )

        table_obj = o.get_table(op.table_name)
        if not table_obj.table_schema.partitions:
            data_srcs = [table_obj]
        elif op.partition is not None and check_partition_exist(
            table_obj, op.partition
        ):
            data_srcs = [table_obj.get_partition(op.partition)]
        else:
            data_srcs = list(table_obj.partitions)
            if op.partition is not None:
                data_srcs = filter_partitions(o, data_srcs, op.partition)

        out_chunks = []
        row_nsplits = []
        index_start = 0
        df = op.outputs[0]

        out_dtypes = df.dtypes
        out_shape = df.shape
        out_columns_value = df.columns_value
        if op.columns is not None:
            out_dtypes = out_dtypes[op.columns]
            out_shape = (df.shape[0], len(op.columns))
            out_columns_value = parse_index(out_dtypes.index, store_data=True)

        for data_src in data_srcs:
            data_store_size = data_src.size
            shape = out_shape
            chunk_size = df.extra_params.chunk_size

            partition_spec = (
                str(data_src.partition_spec)
                if getattr(data_src, "partition_spec", None)
                else None
            )

            if chunk_size is None:
                chunk_bytes = df.extra_params.chunk_bytes or READ_CHUNK_LIMIT
                chunk_count = data_store_size // chunk_bytes + (
                    data_store_size % chunk_bytes != 0
                )
                chunk_size = ceildiv(shape[0], chunk_count)
                split_size = chunk_bytes
            else:
                chunk_count = ceildiv(shape[0], chunk_size)
                split_size = data_store_size // chunk_count

            for i in range(chunk_count):
                start_index = chunk_size * i
                end_index = min(chunk_size * (i + 1), shape[0])
                row_size = end_index - start_index
                chunk_op = DataFrameReadTableSplit(
                    table_name=op.table_name,
                    partition_spec=partition_spec,
                    start_index=start_index,
                    end_index=end_index,
                    nrows=op.nrows,
                    odps_params=op.odps_params,
                    columns=op.columns,
                    add_offset=op.add_offset,
                    dtypes=out_dtypes,
                    sparse=op.sparse,
                    split_size=split_size,
                    use_arrow_dtype=op.use_arrow_dtype,
                    estimate_rows=row_size,
                    append_partitions=op.append_partitions,
                    memory_scale=op.memory_scale,
                    index_columns=op.index_columns,
                )
                index_value = parse_index(pd.RangeIndex(start_index, end_index))
                columns_value = parse_index(out_dtypes.index, store_data=True)
                out_chunk = chunk_op.new_chunk(
                    None,
                    shape=(row_size, out_shape[1]),
                    dtypes=out_dtypes,
                    index_value=index_value,
                    columns_value=columns_value,
                    index=(index_start + i, 0),
                )
                row_nsplits.append(row_size)
                out_chunks.append(out_chunk)

            index_start += chunk_count

        if op.add_offset:
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        nsplits = (tuple(row_nsplits), (out_shape[1],))
        return new_op.new_dataframes(
            None,
            shape=out_shape,
            dtypes=op.dtypes,
            index_value=df.index_value,
            columns_value=out_columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile(cls, op):
        from cupid.runtime import RuntimeContext

        if RuntimeContext.is_context_ready():
            return cls._tile_cupid(op)
        else:
            return cls._tile_tunnel(op)

    if not head_can_be_opt:
        tile = _tile


class DataFrameReadTableSplit(_Base):
    _op_type_ = 123451

    # for cupid
    _cupid_handle = StringField("cupid_handle")
    _split_index = Int64Field("split_index")
    _split_file_start = Int64Field("split_file_start")
    _split_file_end = Int64Field("split_file_end")
    _schema_file_start = Int64Field("schema_file_start")
    _schema_file_end = Int64Field("schema_file_end")
    _use_arrow_dtype = BoolField("use_arrow_dtype")
    _string_as_binary = BoolField("string_as_binary")
    _dtypes = SeriesField("dtypes")
    _nrows = Int64Field("nrows")

    # for tunnel
    _table_name = StringField("table_name")
    _partition_spec = StringField("partition_spec")
    _start_index = Int64Field("start_index")
    _end_index = Int64Field("end_index")
    _odps_params = DictField("odps_params")
    _columns = ListField("columns")

    _split_size = Int64Field("split_size")
    _append_partitions = BoolField("append_partitions")
    _estimate_rows = Int64Field("estimate_rows")
    _meta_raw_size = Int64Field("meta_raw_size")
    _index_columns = ListField("index_columns", ValueType.string)

    def __init__(
        self,
        cupid_handle=None,
        split_index=None,
        split_file_start=None,
        split_file_end=None,
        schema_file_start=None,
        schema_file_end=None,
        table_name=None,
        partition_spec=None,
        start_index=None,
        end_index=None,
        odps_params=None,
        columns=None,
        nrows=None,
        dtypes=None,
        string_as_binary=None,
        split_size=None,
        use_arrow_dtype=None,
        memory_scale=None,
        estimate_rows=None,
        meta_raw_size=None,
        append_partitions=None,
        sparse=None,
        index_columns=None,
        **kw
    ):
        kw.update(_output_type_kw)
        super(DataFrameReadTableSplit, self).__init__(
            _cupid_handle=cupid_handle,
            _split_index=split_index,
            _split_file_start=split_file_start,
            _split_file_end=split_file_end,
            _schema_file_start=schema_file_start,
            _schema_file_end=schema_file_end,
            _table_name=table_name,
            _partition_spec=partition_spec,
            _columns=columns,
            _start_index=start_index,
            _end_index=end_index,
            _odps_params=odps_params,
            _use_arrow_dtype=use_arrow_dtype,
            _string_as_binary=string_as_binary,
            _nrows=nrows,
            _estimate_rows=estimate_rows,
            _split_size=split_size,
            _dtypes=dtypes,
            _append_partitions=append_partitions,
            _sparse=sparse,
            _meta_raw_size=meta_raw_size,
            _memory_scale=memory_scale,
            _index_columns=index_columns,
            **kw
        )

    @property
    def retryable(self):
        return False

    @property
    def output_limit(self):
        return 1

    @property
    def cupid_handle(self):
        return self._cupid_handle

    @property
    def split_index(self):
        return self._split_index

    @property
    def split_file_start(self):
        return self._split_file_start

    @property
    def split_file_end(self):
        return self._split_file_end

    @property
    def schema_file_start(self):
        return self._schema_file_start

    @property
    def schema_file_end(self):
        return self._schema_file_end

    @property
    def table_name(self):
        return self._table_name

    @property
    def partition_spec(self):
        return self._partition_spec

    @property
    def start_index(self):
        return self._start_index

    @property
    def end_index(self):
        return self._end_index

    @property
    def odps_params(self):
        return self._odps_params

    @property
    def columns(self):
        return self._columns

    @property
    def nrows(self):
        return self._nrows

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def split_size(self):
        return self._split_size

    @property
    def estimate_rows(self):
        return self._estimate_rows

    @property
    def use_arrow_dtype(self):
        return self._use_arrow_dtype

    @property
    def string_as_binary(self):
        return self._string_as_binary

    @property
    def append_partitions(self):
        return self._append_partitions

    @property
    def meta_raw_size(self):
        return self._meta_raw_size

    @property
    def index_columns(self):
        return self._index_columns

    @classmethod
    def estimate_size(cls, ctx, op):
        import numpy as np

        def is_object_dtype(dtype):
            try:
                return (
                    np.issubdtype(dtype, np.object_)
                    or np.issubdtype(dtype, np.unicode_)
                    or np.issubdtype(dtype, np.bytes_)
                )
            except TypeError:  # pragma: no cover
                return False

        if op.split_size is None:
            ctx[op.outputs[0].key] = (0, 0)
            return

        arrow_size = (op.memory_scale or ORC_COMPRESSION_RATIO) * op.split_size
        if op.meta_raw_size is not None:
            raw_arrow_size = (op.memory_scale or 1) * op.meta_raw_size
            arrow_size = max(arrow_size, raw_arrow_size)

        n_strings = len([dt for dt in op.dtypes if is_object_dtype(dt)])
        if op.estimate_rows or op.nrows:
            rows = op.nrows if op.nrows is not None else op.estimate_rows
            pd_size = arrow_size + n_strings * rows * STRING_FIELD_OVERHEAD
            logger.debug("Estimate pandas memory cost: %r", pd_size)
        else:
            pd_size = arrow_size * 10 if n_strings else arrow_size

        ctx[op.outputs[0].key] = (pd_size, pd_size + arrow_size)

    @classmethod
    def _cast_string_to_binary(cls, arrow_table):
        import pyarrow as pa

        new_schema = []
        for field in arrow_table.schema:
            if field.type == pa.string():
                new_schema.append(pa.field(field.name, pa.binary()))
            else:
                new_schema.append(field)

        return arrow_table.cast(pa.schema(new_schema))

    @classmethod
    def _append_partition_values(cls, arrow_table, op):
        import pyarrow as pa

        if op.append_partitions and op.partition_spec:
            from odps.types import PartitionSpec

            spec = PartitionSpec(op.partition_spec)

            for col_name, pt_val in spec.items():
                arrow_table = arrow_table.append_column(
                    col_name, pa.array([pt_val] * arrow_table.num_rows, pa.string())
                )

        return arrow_table

    @staticmethod
    def _align_columns(data, expected_dtypes):
        data_columns = data.dtypes.index
        expected_columns = expected_dtypes.index
        if not data_columns.equals(expected_columns):
            logger.debug(
                "Data columns differs from output columns, "
                "data columns: %s, output columns: %s",
                data_columns,
                expected_columns,
            )
            data.columns = expected_columns[: len(data.columns)]
            for extra_col in expected_columns[len(data.columns) :]:
                data[extra_col] = pd.Series([], dtype=expected_dtypes[extra_col])
            if not data.dtypes.index.equals(expected_columns):
                data = data[expected_columns]
        return data

    @classmethod
    def _execute_in_cupid(cls, ctx, op):
        import pyarrow as pa
        from cupid.io.table import TableSplit

        out = op.outputs[0]

        if op.cupid_handle is None:
            empty_df = pd.DataFrame()
            for name, dtype in out.dtypes.items():
                empty_df[name] = pd.Series(dtype=dtype)
            ctx[out.key] = empty_df
            return

        tsp = TableSplit(
            _handle=op.cupid_handle,
            _split_index=op.split_index,
            _split_file_start=op.split_file_start,
            _split_file_end=op.split_file_end,
            _schema_file_start=op.schema_file_start,
            _schema_file_end=op.schema_file_end,
        )
        logger.debug("Read split table, split index: %s", op.split_index)
        reader = tsp.open_arrow_reader()
        if op.nrows is None:
            arrow_table = reader.read_all()
        else:
            nrows = 0
            batches = []
            while nrows < op.nrows:
                try:
                    batch = reader.read_next_batch()
                    nrows += batch.num_rows
                    batches.append(batch)
                except StopIteration:
                    break
            logger.debug("Read %s rows of this split.", op.nrows)
            arrow_table = pa.Table.from_batches(batches)

        arrow_table = cls._append_partition_values(arrow_table, op)

        if op.string_as_binary:
            arrow_table = cls._cast_string_to_binary(arrow_table)
        data = arrow_table_to_pandas_dataframe(
            arrow_table, use_arrow_dtype=op.use_arrow_dtype
        )
        if op.index_columns:
            data = data.set_index(op.index_columns)
        if op.nrows is not None:
            data = data[: op.nrows]

        data = cls._align_columns(data, out.dtypes)

        logger.debug("Read split table finished, split index: %s", op.split_index)
        logger.debug(
            "Split data shape is %s, size is %s",
            data.shape,
            data.memory_usage(deep=True).sum(),
        )
        ctx[out.key] = data

    @classmethod
    def _execute_arrow_tunnel(cls, ctx, op):
        from odps import ODPS
        from odps.tunnel import TableTunnel

        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            odps_params["access_id"],
            odps_params["secret_access_key"],
            project=odps_params["project"],
            endpoint=endpoint,
        )

        t = o.get_table(op.table_name)
        schema_name = t.get_schema().name if t.get_schema() is not None else None
        tunnel = TableTunnel(o, project=t.project)

        if op.partition_spec is not None:
            download_session = tunnel.create_download_session(
                t.name, partition_spec=op.partition_spec
            )
        else:
            download_session = tunnel.create_download_session(t.name, schema=schema_name)
        logger.debug(
            "Start reading table %s(%s) split from %s to %s",
            op.table_name,
            op.partition_spec,
            op.start_index,
            op.end_index,
        )
        if op.nrows is None:
            count = op.end_index - op.start_index
        else:
            count = op.nrows

        columns = op.columns
        if columns is not None:
            columns = (op.index_columns or []) + op.columns
        with download_session.open_arrow_reader(
            op.start_index, count, columns=columns
        ) as reader:
            table = reader.read()

        table = cls._append_partition_values(table, op)
        if op.string_as_binary:
            table = cls._cast_string_to_binary(table)
        data = arrow_table_to_pandas_dataframe(
            table, use_arrow_dtype=op.use_arrow_dtype
        )
        if op.index_columns:
            data = data.set_index(op.index_columns)

        data = cls._align_columns(data, op.outputs[0].dtypes)

        logger.debug(
            "Finish reading table %s(%s) split from %s to %s",
            op.table_name,
            op.partition_spec,
            op.start_index,
            op.end_index,
        )
        ctx[op.outputs[0].key] = data

    @classmethod
    def execute(cls, ctx, op):
        from cupid.runtime import RuntimeContext

        if RuntimeContext.is_context_ready():
            cls._execute_in_cupid(ctx, op)
        else:
            cls._execute_arrow_tunnel(ctx, op)


def df_type_to_np_type(df_type, use_arrow_dtype=False):
    from ....df import types
    from ....df.backends.pd.types import _df_to_np_types

    if df_type == types.string:
        if use_arrow_dtype:
            return ArrowStringDtype()
        else:
            return np.dtype("object")
    elif df_type in _df_to_np_types:
        return _df_to_np_types[df_type]
    elif df_type == types.timestamp:
        return np.datetime64(0, "ns").dtype
    else:
        return np.dtype("object")


def read_odps_table(
    table,
    shape,
    partition=None,
    sparse=False,
    chunk_bytes=None,
    chunk_size=None,
    columns=None,
    odps_params=None,
    add_offset=False,
    use_arrow_dtype=False,
    string_as_binary=None,
    memory_scale=None,
    append_partitions=False,
    with_split_meta_on_tile=False,
    index_columns=None,
):
    import pandas as pd

    if isinstance(chunk_size, (list, tuple)):
        if len(chunk_size) == 1:
            chunk_size = chunk_size[0]
        if len(chunk_size) > 1:
            raise ValueError("Only support split on rows")

    if chunk_bytes is not None:
        chunk_bytes = int(parse_readable_size(chunk_bytes)[0])
    cols = table.table_schema.columns if append_partitions else table.table_schema.simple_columns
    table_columns = [c.name for c in cols]
    table_types = [c.type for c in cols]
    df_types = [
        df_type_to_np_type(odps_type_to_df_type(type), use_arrow_dtype=use_arrow_dtype)
        for type in table_types
    ]

    if isinstance(index_columns, str):
        index_columns = [index_columns]
    if index_columns and columns is None:
        index_col_set = set(index_columns)
        columns = [c for c in table_columns if c not in index_col_set]

    if not index_columns:
        index_dtypes = None
    else:
        table_index_types = [
            df_types[table_columns.index(col)] for col in index_columns
        ]
        index_dtypes = pd.Series(table_index_types, index=index_columns)

    if columns is not None:
        table_col_set = set(columns)
        if any(col in table_col_set for col in index_columns):
            raise ValueError("Index columns and columns shall not overlap.")

        # reorder columns
        new_columns = [c for c in table_columns if c in columns]
        df_types = [df_types[table_columns.index(col)] for col in new_columns]
        table_columns = new_columns
        columns = new_columns

    dtypes = pd.Series(df_types, index=table_columns)

    op = DataFrameReadTable(
        odps_params=odps_params,
        table_name=table.full_table_name,
        partition_spec=partition,
        dtypes=dtypes,
        sparse=sparse,
        add_offset=add_offset,
        columns=columns,
        use_arrow_dtype=use_arrow_dtype,
        string_as_binary=string_as_binary,
        memory_scale=memory_scale,
        append_partitions=append_partitions,
        last_modified_time=to_timestamp(table.last_data_modified_time),
        with_split_meta_on_tile=with_split_meta_on_tile,
        index_columns=index_columns,
        index_dtypes=index_dtypes,
    )
    return op(shape, chunk_bytes=chunk_bytes, chunk_size=chunk_size)


class ReadODPSTableHeadRule(DataSourceHeadRule):
    @staticmethod
    def match(chunk, graph, keys):
        from mars.dataframe.indexing.iloc import DataFrameIlocGetItem, SeriesIlocGetItem

        op = chunk.op
        inputs = graph.predecessors(chunk)
        if (
            len(inputs) == 1
            and isinstance(op, (DataFrameIlocGetItem, SeriesIlocGetItem))
            and isinstance(inputs[0].op, DataFrameReadTableSplit)
            and inputs[0].key not in keys
        ):
            try:
                is_head = op.can_be_optimized()
            except AttributeError:
                is_head = op.is_head()
            if is_head:
                return True
            else:
                return False
        return False
